# coding: utf-8

# ------------------------------------------------------------------------------
# Copyright (C) 2019 Maximilian Stahlberg
# Based on the original picos.expressions module by Guillaume Sagnol.
#
# This file is part of PICOS.
#
# PICOS is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PICOS is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# ------------------------------------------------------------------------------

"""Implements :class:`Entropy` and :class:`NegativeEntropy`."""

import math
import operator
from collections import namedtuple

import cvxopt
import numpy

from .. import glyphs
from ..apidoc import api_end, api_start
from ..caching import cached_selfinverse_unary_operator, cached_unary_operator
from ..constraints import KullbackLeiblerConstraint
from .data import convert_and_refine_arguments, convert_operands, cvx2np
from .exp_affine import AffineExpression
from .expression import Expression, refine_operands, validate_prediction

_API_START = api_start(globals())
# -------------------------------


class Entropy(Expression):
    r"""Entropy or negative relative entropy of an affine expression.

    Negative relative entropy is also known as the perspective of the logarithm.

    :Definition:

    Let :math:`x` be an :math:`n`-dimensional real affine expression.

    1.  If no additional expression :math:`y` is given, this is the entropy

        .. math::

            -\sum_{i = 1}^n \operatorname{vec}(x)_i
            \log(\operatorname{vec}(x)_i).

    2.  If an additional affine expression :math:`y` of same shape as :math:`x`
        is given, this is the negative relative entropy (or logarithmic
        perspective)

        .. math::

            -\sum_{i = 1}^n \operatorname{vec}(x)_i
            \log\left(
                \frac{\operatorname{vec}(x)_i}{\operatorname{vec}(y)_i}
            \right)
            &= -\sum_{i = 1}^n \operatorname{vec}(x)_i
            \left[
                \log(\operatorname{vec}(x)_i) - \log(\operatorname{vec}(y)_i)
            \right] \\
            &= \sum_{i = 1}^n \operatorname{vec}(x)_i
            \left[
                \log(\operatorname{vec}(y)_i) - \log(\operatorname{vec}(x)_i
            \right] \\
            &= \sum_{i = 1}^n \operatorname{vec}(x)_i
            \log\left(
                \frac{\operatorname{vec}(y)_i}{\operatorname{vec}(x)_i}
            \right).

    .. warning::

        When you pose a lower bound on this expression, then PICOS enforces
        :math:`x \geq 0` through an auxiliary constraint during solution search.
        When an additional expression :math:`y` is given, PICOS enforces
        :math:`y \geq 0` as well.
    """

    # --------------------------------------------------------------------------
    # Initialization and factory methods.
    # --------------------------------------------------------------------------

    @convert_and_refine_arguments("x", "y", allowNone=True)
    def __init__(self, x, y=None):
        """Construct an :class:`Entropy`.

        :param x: The affine expression :math:`x`.
        :type x: ~picos.expressions.AffineExpression
        :param y: An additional affine expression :math:`y`. If necessary, PICOS
            will attempt to reshape or broadcast it to the shape of :math:`x`.
        :type y: ~picos.expressions.AffineExpression
        """
        if not isinstance(x, AffineExpression):
            raise TypeError("Can only take the elementwise logarithm of a real "
                "affine expression, not of {}.".format(type(x).__name__))

        if y is not None:
            if not isinstance(y, AffineExpression):
                raise TypeError("The additional parameter y must be a real "
                    "affine expression, not {}.".format(type(x).__name__))
            elif x.shape != y.shape:
                y = y.reshaped_or_broadcasted(x.shape)

            if y.is1:
                y = None

        self._x = x
        self._y = y

        if y is None:
            typeStr = "Entropy"
            if len(x) == 1:
                symbStr = glyphs.neg(glyphs.mul(x.string, glyphs.log(x.string)))
            else:
                symbStr = glyphs.neg(
                    glyphs.dotp(x.string, glyphs.log(x.string)))
        else:
            typeStr = "Logarithmic Perspective"
            if len(x) == 1:
                symbStr = glyphs.mul(
                    x.string, glyphs.log(glyphs.div(y.string, x.string)))
            else:
                symbStr = glyphs.dotp(x.string,
                    glyphs.sub(glyphs.log(y.string), glyphs.log(x.string)))

        Expression.__init__(self, typeStr, symbStr)

    # --------------------------------------------------------------------------
    # Abstract method implementations and method overridings, except _predict.
    # --------------------------------------------------------------------------

    def _get_refined(self):
        if self._x.constant and (self._y is None or self._y.constant):
            return AffineExpression.from_constant(self.value, 1, self._symbStr)
        else:
            return self

    Subtype = namedtuple("Subtype", ("argdim", "y"))

    def _get_subtype(self):
        return self.Subtype(len(self._x), self._y is not None)

    def _get_value(self):
        x = numpy.ravel(cvx2np(self._x._get_value()))

        if self._y is None:
            s = -numpy.dot(x, numpy.log(x))
        else:
            y = numpy.ravel(cvx2np(self._y._get_value()))
            s = numpy.dot(x, numpy.log(y / x))

        return cvxopt.matrix(s)

    @cached_unary_operator
    def _get_variables(self):
        if self._y is None:
            return self._x.variables
        else:
            return self._x.variables.union(self._y.variables)

    def _is_convex(self):
        return False

    def _is_concave(self):
        return True

    def _replace_variables(self, var_map):
        return self.__class__(self._x._replace_variables(var_map),
            None if self._y is None else self._y._replace_variables(var_map))

    # --------------------------------------------------------------------------
    # Python special method implementations, except constraint-creating ones.
    # --------------------------------------------------------------------------

    @cached_selfinverse_unary_operator
    def __neg__(self):
        return NegativeEntropy(self._x, self._y)

    @convert_operands(scalarRHS=True)
    @refine_operands()
    def __add__(self, other):
        if isinstance(other, AffineExpression):
            other = Entropy(other, other * math.e)

        if isinstance(other, Entropy):
            if self._y is None and other._y is None:
                entropy = Entropy(self._x.vec // other._x.vec)
            elif self._y is not None and other._y is None:
                one = AffineExpression.from_constant(1.0, (other.n, 1))
                entropy = Entropy(
                    self._x.vec // other._x.vec, self._y.vec // one)
            elif self._y is None and other._y is not None:
                one = AffineExpression.from_constant(1.0, (self.n, 1))
                entropy = Entropy(
                    self._x.vec // other._x.vec, one // other._y.vec)
            else:
                entropy = Entropy(
                    self._x.vec // other._x.vec, self._y.vec // other._y.vec)

            entropy._symbStr = glyphs.clever_add(self.string, other.string)

            return entropy
        else:
            return NotImplemented

    @convert_operands(scalarRHS=True)
    @refine_operands()
    def __sub__(self, other):
        if isinstance(other, AffineExpression):
            other = NegativeEntropy(other, other / math.e)

        if isinstance(other, NegativeEntropy):
            return self + (-other)
        else:
            return NotImplemented

    @convert_operands(scalarRHS=True)
    @refine_operands()
    def __radd__(self, other):
        if isinstance(other, (AffineExpression, Entropy)):
            entropy = self.__add__(other)
            # NOTE: __add__ always creates a fresh expression.
            entropy._symbStr = glyphs.clever_add(other.string, self.string)
            return entropy
        else:
            return NotImplemented

    # --------------------------------------------------------------------------
    # Methods and properties that return expressions.
    # --------------------------------------------------------------------------

    @property
    def x(self):
        """The expression :math:`x`."""
        return self._x

    @property
    def y(self):
        """The additional expression :math:`y`, or :obj:`None`."""
        return self._y

    # --------------------------------------------------------------------------
    # Methods and properties that describe the expression.
    # --------------------------------------------------------------------------

    @property
    def n(self):
        """Length of :attr:`x`."""
        return len(self._x)

    # --------------------------------------------------------------------------
    # Constraint-creating operators, and _predict.
    # --------------------------------------------------------------------------

    @classmethod
    def _predict(cls, subtype, relation, other):
        assert isinstance(subtype, cls.Subtype)

        if relation == operator.__ge__:
            if issubclass(other.clstype, AffineExpression) \
            and other.subtype.dim == 1:
                return KullbackLeiblerConstraint.make_type(
                    argdim=subtype.argdim)

        return NotImplemented

    @convert_operands(scalarRHS=True)
    @validate_prediction
    @refine_operands()
    def __ge__(self, other):
        if isinstance(other, AffineExpression):
            return KullbackLeiblerConstraint(-self, -other)
        else:
            return NotImplemented


class NegativeEntropy(Expression):
    r"""Negative or relative entropy of an affine expression.

    Relative entropy is also known as the Kullback-Leibler divergence.

    :Definition:

    Let :math:`x` be an :math:`n`-dimensional real affine expression.

    1.  If no additional expression :math:`y` is given, this is the negative
        entropy

        .. math::

            \sum_{i = 1}^n \operatorname{vec}(x)_i
            \log(\operatorname{vec}(x)_i).

    2.  If an additional affine expression :math:`y` of same shape as :math:`x`
        is given, this is the relative entropy (or Kullback-Leibler divergence)

        .. math::

            \sum_{i = 1}^n \operatorname{vec}(x)_i
            \log\left(
                \frac{\operatorname{vec}(x)_i}{\operatorname{vec}(y)_i}
            \right)
            = \sum_{i = 1}^n \operatorname{vec}(x)_i
            \left[
                \log(\operatorname{vec}(x)_i) - \log(\operatorname{vec}(y)_i)
            \right].

    .. warning::

        When you pose an upper bound on this expression, then PICOS enforces
        :math:`x \geq 0` through an auxiliary constraint during solution search.
        When an additional expression :math:`y` is given, PICOS enforces
        :math:`y \geq 0` as well.
    """

    # --------------------------------------------------------------------------
    # Initialization and factory methods.
    # --------------------------------------------------------------------------

    @convert_and_refine_arguments("x", "y", allowNone=True)
    def __init__(self, x, y=None):
        """Construct a :class:`NegativeEntropy`.

        :param x: The affine expression :math:`x`.
        :type x: ~picos.expressions.AffineExpression
        :param y: An additional affine expression :math:`y`. If necessary, PICOS
            will attempt to reshape or broadcast it to the shape of :math:`x`.
        :type y: ~picos.expressions.AffineExpression
        """
        if not isinstance(x, AffineExpression):
            raise TypeError("Can only take the elementwise logarithm of a real "
                "affine expression, not of {}.".format(type(x).__name__))

        if y is not None:
            if not isinstance(y, AffineExpression):
                raise TypeError("The additional parameter y must be a real "
                    "affine expression, not {}.".format(type(x).__name__))
            elif x.shape != y.shape:
                y = y.reshaped_or_broadcasted(x.shape)

            if y.is1:
                y = None

        self._x = x
        self._y = y

        if y is None:
            typeStr = "Negative Entropy"
            if len(x) == 1:
                symbStr = glyphs.mul(x.string, glyphs.log(x.string))
            else:
                symbStr = glyphs.dotp(x.string, glyphs.log(x.string))
        else:
            typeStr = "Relative Entropy"
            if len(x) == 1:
                symbStr = glyphs.mul(
                    x.string, glyphs.log(glyphs.div(x.string, y.string)))
            else:
                symbStr = glyphs.dotp(x.string,
                    glyphs.sub(glyphs.log(x.string), glyphs.log(y.string)))

        Expression.__init__(self, typeStr, symbStr)

    # --------------------------------------------------------------------------
    # Abstract method implementations and method overridings, except _predict.
    # --------------------------------------------------------------------------

    def _get_refined(self):
        if self._x.constant and (self._y is None or self._y.constant):
            return AffineExpression.from_constant(self.value, 1, self._symbStr)
        else:
            return self

    Subtype = namedtuple("Subtype", ("argdim", "y"))

    def _get_subtype(self):
        return self.Subtype(len(self._x), self._y is not None)

    def _get_value(self):
        x = numpy.ravel(cvx2np(self._x._get_value()))

        if self._y is None:
            s = numpy.dot(x, numpy.log(x))
        else:
            y = numpy.ravel(cvx2np(self._y._get_value()))
            s = numpy.dot(x, numpy.log(x / y))

        return cvxopt.matrix(s)

    @cached_unary_operator
    def _get_variables(self):
        if self._y is None:
            return self._x.variables
        else:
            return self._x.variables.union(self._y.variables)

    def _is_convex(self):
        return True

    def _is_concave(self):
        return False

    def _replace_variables(self, var_map):
        return self.__class__(self._x._replace_variables(var_map),
            None if self._y is None else self._y._replace_variables(var_map))

    # --------------------------------------------------------------------------
    # Python special method implementations, except constraint-creating ones.
    # --------------------------------------------------------------------------

    @cached_selfinverse_unary_operator
    def __neg__(self):
        return Entropy(self._x, self._y)

    @convert_operands(scalarRHS=True)
    @refine_operands()
    def __add__(self, other):
        if isinstance(other, AffineExpression):
            other = NegativeEntropy(other, other / math.e)

        if isinstance(other, NegativeEntropy):
            if self._y is None and other._y is None:
                negent = NegativeEntropy(self._x.vec // other._x.vec)
            elif self._y is not None and other._y is None:
                one = AffineExpression.from_constant(1.0, (other.n, 1))
                negent = NegativeEntropy(
                    self._x.vec // other._x.vec, self._y.vec // one)
            elif self._y is None and other._y is not None:
                one = AffineExpression.from_constant(1.0, (self.n, 1))
                negent = NegativeEntropy(
                    self._x.vec // other._x.vec, one // other._y.vec)
            else:
                negent = NegativeEntropy(
                    self._x.vec // other._x.vec, self._y.vec // other._y.vec)

            negent._symbStr = glyphs.clever_add(self.string, other.string)

            return negent
        else:
            return NotImplemented

    @convert_operands(scalarRHS=True)
    @refine_operands()
    def __radd__(self, other):
        if isinstance(other, (AffineExpression, NegativeEntropy)):
            negent = self.__add__(other)
            # NOTE: __add__ always creates a fresh expression.
            negent._symbStr = glyphs.clever_add(other.string, self.string)
            return negent
        else:
            return NotImplemented

    @convert_operands(scalarRHS=True)
    @refine_operands()
    def __sub__(self, other):
        if isinstance(other, AffineExpression):
            other = Entropy(other, other * math.e)

        if isinstance(other, Entropy):
            return self + (-other)
        else:
            return NotImplemented

    # --------------------------------------------------------------------------
    # Methods and properties that return expressions.
    # --------------------------------------------------------------------------

    @property
    def x(self):
        """The expression :math:`x`."""
        return self._x

    @property
    def y(self):
        """The additional expression :math:`y`, or :obj:`None`."""
        return self._y

    # --------------------------------------------------------------------------
    # Methods and properties that describe the expression.
    # --------------------------------------------------------------------------

    @property
    def n(self):
        """Length of :attr:`x`."""
        return len(self._x)

    # --------------------------------------------------------------------------
    # Constraint-creating operators, and _predict.
    # --------------------------------------------------------------------------

    @classmethod
    def _predict(cls, subtype, relation, other):
        assert isinstance(subtype, cls.Subtype)

        if relation == operator.__le__:
            if issubclass(other.clstype, AffineExpression) \
            and other.subtype.dim == 1:
                return KullbackLeiblerConstraint.make_type(
                    argdim=subtype.argdim)

        return NotImplemented

    @convert_operands(scalarRHS=True)
    @validate_prediction
    @refine_operands()
    def __le__(self, other):
        if isinstance(other, AffineExpression):
            return KullbackLeiblerConstraint(self, other)
        else:
            return NotImplemented


# --------------------------------------
__all__ = api_end(_API_START, globals())



# coding: utf-8

# ------------------------------------------------------------------------------
# Copyright (C) 2020 Maximilian Stahlberg
#
# This file is part of PICOS.
#
# PICOS is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PICOS is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# ------------------------------------------------------------------------------

"""Backwards compatibility version of an older module."""

from .apidoc import api_end, api_start
from .expressions.data import load_data
from .legacy import deprecated, throw_deprecation_warning

# Deprecated as of 2.0.
throw_deprecation_warning("The 'picos.tools' module is deprecated and will be "
    "removed in a future release. Try importing from 'picos' or 'picos.algebra'"
    "instead.")

_API_START = api_start(globals())
# -------------------------------


from builtins import sum as builtin_sum  # noqa

from . import (  # noqa
    sum, geomean, norm, tracepow, trace, sum_k_largest, sum_k_largest_lambda,
    lambda_max, sum_k_smallest, sum_k_smallest_lambda, lambda_min,
    partial_transpose, partial_trace, detrootn, ball, simplex,
    truncated_simplex, expcone, exp, log, sumexp, kullback_leibler, logsumexp,
    lse, diag, diag_vect, new_param, import_cbf, kron, flow_Constraint
)

from .formatting import (  # noqa
    detect_range, parameterized_string
)


@deprecated("2.0", reason="Write '{k: v.value for k, v in x.items()}' instead.")
def eval_dict(dict_of_variables):
    """Evaluate all values of a dictionary.

    :param dict dict_of_variables:
        A dictionary mapping arbitrary keys to PICOS objects that have a
        ``value`` attribute, such as variables.

    :returns:
        Another dictionary with the original dicitonary's values replaced by the
        value of their ``value`` attribute.
    """
    return {k: v.value for k, v in dict_of_variables.items()}


@deprecated("2.0", useInstead="picos.expressions.data.load_data")
def retrieve_matrix(mat, exSize=None):
    """Legacy wrapper around :func:`~.expressions.data.load_data`."""
    return load_data(value=mat, shape=exSize, sparse=True, legacy=True)


# --------------------------------------
__all__ = api_end(_API_START, globals())



# coding: utf-8

# ------------------------------------------------------------------------------
# Copyright (C) 2019 Maximilian Stahlberg
#
# This file is part of PICOS.
#
# PICOS is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PICOS is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# ------------------------------------------------------------------------------

"""Caching helpers."""

import contextlib
import functools

from .apidoc import api_end, api_start

_API_START = api_start(globals())
# -------------------------------


#: The prefix used for storing cached values.
CACHED_PREFIX = "_cached_"

#: An attribute name whose presence unlocks the setter of cached properties.
CACHED_PROP_UNLOCKED_TOKEN = "_CACHED_PROPERTIES_ARE_UNLOCKED"


@contextlib.contextmanager
def unlocked_cached_properties(obj):
    """Unlock the setters of cached instance attributes.

    Normally, cached attributes are read-only properties. When the user first
    reads them, the cache is populated with the value returned to the user, and
    successive reads will return the cached value.

    The user is allowed to empty the cache by using ``del`` on the variable, but
    they may not assign a value to it. This context allows the programmer to
    manually populate the cache by assigning a value to the property.

    :Example:

    >>> from picos.caching import cached_property, unlocked_cached_properties
    >>> class A:
    ...     @cached_property
    ...     def p(self):
    ...         return 1
    ...
    >>> a = A()
    >>> try:
    ...     a.p = 2
    ... except AttributeError:
    ...     print("Not possible.")
    ...
    Not possible.
    >>> with unlocked_cached_properties(a):
    ...     a.p = 2  # Populate the cache of a.p.
    ...
    >>> a.p
    2
    """
    assert not hasattr(obj, CACHED_PROP_UNLOCKED_TOKEN)
    setattr(obj, CACHED_PROP_UNLOCKED_TOKEN, None)

    try:
        yield
    finally:
        delattr(obj, CACHED_PROP_UNLOCKED_TOKEN)


class cached_property(property):
    """A read-only property whose result is cached."""

    def __init__(self, fget=None, fset=None, fdel=None, doc=None):  # noqa
        if fget is None:
            raise NotImplementedError("Unlike normal properties, cached "
                "properties must be initialized with a getter.")
        elif fget.__name__ == (lambda: None).__name__:
            raise NotImplementedError("Cached properties cannot be used with a "
                "lambda getter as lambdas do not have a unique name to identify"
                " the cached value.")

        if fset is not None:
            raise AttributeError("Cannot have a custom setter on a cached "
                "property as __set__ is reserved for cache population.")

        if fdel is not None:
            raise AttributeError("Cannot have a custom deleter on a cached "
                "property as __delete__ is reserved for cache clearing.")

        self._cache_name = CACHED_PREFIX + fget.__name__

        property.__init__(self, fget, fset, fdel, doc)

    def __get__(self, obj, objtype=None):
        if obj is None:
            return self

        if self.fget is None:
            raise AttributeError("unreadable attribute")  # Mimic Python.

        if not hasattr(obj, self._cache_name):
            # Compute and cache the value.
            setattr(obj, self._cache_name, self.fget(obj))

        return getattr(obj, self._cache_name)

    def __set__(self, obj, value):
        if not hasattr(obj, CACHED_PROP_UNLOCKED_TOKEN):
            raise AttributeError("can't set attribute")  # Mimic Python.

        # Populate the cache.
        setattr(obj, self._cache_name, value)

    def __delete__(self, obj):
        # Empty the cache.
        if hasattr(obj, self._cache_name):
            delattr(obj, self._cache_name)

    def getter(self, fget):  # noqa
        raise NotImplementedError(
            "Cannot change the getter on a cached property.")

    def setter(self, fset):  # noqa
        raise AttributeError("Cannot add a setter to a cached property.")

    def deleter(self, fdel):  # noqa
        return AttributeError("Cannot add a deleter to a cached property.")


class cached_selfinverse_property(cached_property):
    """A read-only, self-inverse property whose result is cached."""

    def __get__(self, obj, objtype=None):
        if obj is None:
            return self

        if self.fget is None:
            raise AttributeError("unreadable attribute")  # Mimic Python.

        if not hasattr(obj, self._cache_name):
            # Compute and cache the value, populate the value's cache.
            value = self.fget(obj)
            setattr(obj, self._cache_name, value)
            setattr(value, self._cache_name, obj)

        return getattr(obj, self._cache_name)


def cached_unary_operator(operator):
    """Make a unary operator method cache its result.

    This is supposed to be used for property-like special methods such as
    ``__neg__`` where :func:`cached_property` can't be used.
    """
    cacheName = CACHED_PREFIX + operator.__name__

    @functools.wraps(operator)
    def wrapper(self):
        if not hasattr(self, cacheName):
            setattr(self, cacheName, operator(self))
        return getattr(self, cacheName)
    return wrapper


def cached_selfinverse_unary_operator(operator):
    """Make a self-inverse unary operator method cache its result.

    This is supposed to be used for property-like special methods such as
    ``__neg__`` where :func:`cached_property` can't be used.

    .. warning::
        The result returned by the wrapped operator must be a fresh object as
        it will be modified.
    """
    cacheName = CACHED_PREFIX + operator.__name__

    @functools.wraps(operator)
    def wrapper(self):
        if not hasattr(self, cacheName):
            value = operator(self)
            setattr(self, cacheName, value)
            setattr(value, cacheName, self)
        return getattr(self, cacheName)
    return wrapper


def empty_cache(obj):
    """Clear all cached values of an object."""
    for name in dir(obj):
        if name.startswith(CACHED_PREFIX):
            delattr(obj, name)


def borrow_cache(target, source, names):
    """Copy cached values from one object to another.

    :param target: The object to populate the cache of.
    :param source: The object to take cached values from.
    :param names: Names of cached properties or functions to borrow.
    """
    if not isinstance(target, type(source)):
        raise TypeError(
            "The target object must be of a subclass of the source type.")

    for name in names:
        cacheName = CACHED_PREFIX + name
        if hasattr(source, cacheName):
            setattr(target, cacheName, getattr(source, cacheName))


# --------------------------------------
__all__ = api_end(_API_START, globals())



# coding: utf-8

# ------------------------------------------------------------------------------
# Copyright (C) 2019 Maximilian Stahlberg
# Based on the original picos.expressions module by Guillaume Sagnol.
#
# This file is part of PICOS.
#
# PICOS is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PICOS is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# ------------------------------------------------------------------------------

"""Implements :class:`Logarithm`."""

import math
import operator
from collections import namedtuple

import cvxopt

from .. import glyphs
from ..apidoc import api_end, api_start
from ..caching import cached_unary_operator
from ..constraints import LogConstraint
from .data import convert_and_refine_arguments, convert_operands
from .exp_affine import AffineExpression
from .expression import Expression, refine_operands, validate_prediction

_API_START = api_start(globals())
# -------------------------------


class Logarithm(Expression):
    r"""Logarithm of a scalar affine expression.

    :Definition:

    For a real scalar affine expression :math:`x`, this is :math:`\log(x)`.

    .. warning::

        When you pose a lower bound on a logarithm :math:`\log(x)`, then PICOS
        enforces :math:`x \geq 0` through an auxiliary constraint during
        solution search.
    """

    # --------------------------------------------------------------------------
    # Initialization and factory methods.
    # --------------------------------------------------------------------------

    @convert_and_refine_arguments("x")
    def __init__(self, x):
        """Construct a :class:`Logarithm`.

        :param x: The scalar affine expression :math:`x`.
        :type x: ~picos.expressions.AffineExpression
        """
        if not isinstance(x, AffineExpression):
            raise TypeError("Can only take the logarithm of a real affine "
                "expression, not of {}.".format(type(x).__name__))
        elif not x.scalar:
            raise TypeError("Can only take the logarithm of a scalar expression"
                "but {} is shaped {}.".format(x.string, glyphs.shape(x.shape)))

        self._x = x

        Expression.__init__(self, "Logarithm", glyphs.log(x.string))

    # --------------------------------------------------------------------------
    # Abstract method implementations and method overridings, except _predict.
    # --------------------------------------------------------------------------

    def _get_refined(self):
        if self._x.constant:
            return AffineExpression.from_constant(self.value, 1, self._symbStr)
        else:
            return self

    Subtype = namedtuple("Subtype", ())

    def _get_subtype(self):
        return self.Subtype()

    def _get_value(self):
        value = cvxopt.matrix(self._x._get_value())  # Must be dense for log.
        return cvxopt.log(value)

    @cached_unary_operator
    def _get_variables(self):
        return self._x.variables

    def _is_convex(self):
        return False

    def _is_concave(self):
        return True

    def _replace_variables(self, var_map):
        return self.__class__(self._x._replace_variables(var_map))

    # --------------------------------------------------------------------------
    # Python special method implementations, except constraint-creating ones.
    # --------------------------------------------------------------------------

    @convert_operands(scalarRHS=True)
    @refine_operands()
    def __add__(self, other):
        if isinstance(other, AffineExpression):
            if not other.constant:
                raise NotImplementedError("You may only add a constant term to "
                    "a nonconstant PICOS logarithm.")

            log = Logarithm(self._x * math.exp(other.value))
            log._typeStr = "Offset " + log._typeStr
            log._symbStr = glyphs.clever_add(self.string, other.string)

            return log

    @convert_operands(scalarRHS=True)
    @refine_operands()
    def __radd__(self, other):
        if isinstance(other, AffineExpression):
            log = self.__add__(other)
            # NOTE: __add__ always creates a fresh expression.
            log._symbStr = glyphs.clever_add(other.string, self.string)
            return log
        else:
            return NotImplemented

    @convert_operands(scalarRHS=True)
    @refine_operands()
    def __sub__(self, other):
        if isinstance(other, AffineExpression):
            if not other.constant:
                raise NotImplementedError("You may only substract a constant "
                    "term from a nonconstant PICOS logarithm.")

            log = Logarithm(self._x / math.exp(other.value))
            log._typeStr = "Offset " + log._typeStr
            log._symbStr = glyphs.clever_sub(self.string, other.string)

            return log

    @convert_operands(scalarRHS=True)
    @refine_operands()
    def __mul__(self, other):
        from . import Entropy, NegativeEntropy

        if isinstance(other, AffineExpression):
            if other.is0:
                return AffineExpression.zero()
            elif other.is1:
                # NOTE: We could return self here, but this is more consistent
                #       with other expressions' __mul__ methods.
                return Logarithm(self._x)
            elif other.equals(self._x):
                return NegativeEntropy(self._x)
            elif other.equals(-self._x):
                return Entropy(self._x)
            else:
                raise NotImplementedError(
                    "You may multiply {} only with {}, 0, or 1."
                    .format(self.string, glyphs.plsmns(self._x.string)))
        else:
            return NotImplemented

    @convert_operands(scalarRHS=True)
    @refine_operands()
    def __rmul__(self, other):
        if isinstance(other, AffineExpression):
            return self.__mul__(other)
        else:
            return NotImplemented

    # --------------------------------------------------------------------------
    # Methods and properties that return expressions.
    # --------------------------------------------------------------------------

    @property
    def x(self):
        """The expression :math:`x`."""
        return self._x

    @property
    def exp(self):
        """The exponential of the logarithm, equal to :math:`x`."""
        return self._x

    # --------------------------------------------------------------------------
    # Constraint-creating operators, and _predict.
    # --------------------------------------------------------------------------

    @classmethod
    def _predict(cls, subtype, relation, other):
        assert isinstance(subtype, cls.Subtype)

        if relation == operator.__ge__:
            if issubclass(other.clstype, AffineExpression) \
            and other.subtype.dim == 1:
                return LogConstraint.make_type()

        return NotImplemented

    @convert_operands(scalarRHS=True)
    @validate_prediction
    @refine_operands()
    def __ge__(self, other):
        if isinstance(other, AffineExpression):
            return LogConstraint(self, other)
        else:
            return NotImplemented


# --------------------------------------
__all__ = api_end(_API_START, globals())



# coding: utf-8

# ------------------------------------------------------------------------------
# Copyright (C) 2019-2020 Maximilian Stahlberg
#
# This file is part of PICOS.
#
# PICOS is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PICOS is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# ------------------------------------------------------------------------------

"""Optimization solver parameter handling."""

# ------------------------------------------------------------------------------
# NOTE: When modifying tolerance options, be sure to also modify tolerances.rst.
# ------------------------------------------------------------------------------

import fnmatch
import types

from ..apidoc import api_end, api_start
from ..solvers import all_solvers

_API_START = api_start(globals())
# -------------------------------


OPTIONS = [
    # General options.
    # --------------------------------------------------------------------------

    ("strict_options", bool, False, """
        Whether unsupported general options will raise an
        :class:`UnsupportedOptionError
        <picos.solvers.solver.UnsupportedOptionError>`
        exception, instead of printing a warning."""),

    ("verbosity", int, 0, """
        Verbosity level.

        - ``-1`` attempts to suppress all output, even errros.
        - ``0`` only generates warnings and errors.
        - ``1`` generates standard informative output.
        - ``2`` or larger prints additional information for debugging purposes.
        """, lambda n: -1 <= n),

    ("license_warnings", bool, True, """
        Whether solvers are allowed to ignore the :ref:`verbosity
        <option_verbosity>` option to print licensing related warnings.

        .. warning::

            Using this option to suppress licensing related warnings is done at
            your own legal responsibility."""),

    ("solver", str, None, """
        The solver to use.

        - :obj:`None` to let PICOS choose.
        - """ + """
        - """.join('``"{0}"`` for :class:`{1} <picos.solvers.{1}>`.'
            .format(name, solver.__name__)
            for name, solver in all_solvers().items()) + """

        .. note::

            :func:`picos.available_solvers() <picos.available_solvers>` returns
            a list of names of solvers that are available at runtime.
        """, lambda value: value is None or value in all_solvers().keys()),

    ("primals", bool, True, """
        Whether to request a primal solution.

        - :obj:`True` will raise an exception if no optimal primal solution is
          found.
        - :obj:`None` will accept and apply also incomplete, infeasible or
          suboptimal primal solutions.
        - :obj:`False` will not ask for a primal solution and throw away any
          primal solution returned by the solver.
        """, None),

    ("duals", bool, None, """
        Whether to request a dual solution.

        - :obj:`True` will raise an exception if no optimal dual solution is
          found.
        - :obj:`None` will accept and apply also incomplete, infeasible or
          suboptimal dual solutions.
        - :obj:`False` will not ask for a dual solution and throw away any
          dual solution returned by the solver.
        """, None),

    ("dualize", bool, False, """
        Whether to dualize the problem as part of the solution strategy.

        This can sometimes lead to a significant solution search speedup.
     """),

    ("assume_conic", bool, True, r"""
        Determines how :class:`~picos.constraints.ConicQuadraticConstraint`
        instances, which correspond to nonconvex constraints of the form
        :math:`x^TQx + p^Tx + q \leq (a^Tx + b)(c^Tx + d)` with
        :math:`x^TQx + p^Tx + q` representable as a squared norm, are processed:

        - :obj:`True` strengthens them into convex conic constraints by assuming
          the additional constraints :math:`a^Tx + b \geq 0` and
          :math:`c^Tx + d \geq 0`.
        - :obj:`False` takes them verbatim and also considers solutions with
          :math:`(a^Tx + b) < 0` or :math:`(c^Tx + d) < 0`. This requires a
          solver that accepts nonconvex quadratic constraints.

        .. warning::

            :class:`~picos.constraints.ConicQuadraticConstraint` are also used
            in the case of :math:`Q = 0`. For instance, :math:`x^2 \geq 1` is
            effectively ransformed to :math:`x \geq 1` if this is :obj:`True`.
        """),

    ("apply_solution", bool, True, """
        Whether to immediately apply the solution returned by a solver to the
        problem's variables and constraints.

        If multiple solutions are returned by the solver, then the first one
        will be applied. If this is ``False``, then solutions can be applied
        manually via their :func:`apply <picos.Solution.apply>` method. """),

    ("abs_prim_fsb_tol", float, 1e-8, """
        Absolute primal problem feasibility tolerance.

        A primal solution is feasible if some norm over the vector of primal
        constraint violations is smaller than this value.

        :obj:`None` lets the solver use its own default value.
        """, lambda tol: tol is None or tol > 0.0),

    ("rel_prim_fsb_tol", float, 1e-8, """
        Relative primal problem feasibility tolerance.

        Like :ref:`abs_prim_fsb_tol <option_abs_prim_fsb_tol>`, but the norm is
        divided by the norm of the constraints' right hand side vector.

        If the norm used is some nested norm (e.g. the maximum over the norms of
        the equality and inequality violations), then solvers might divide the
        inner violation norms by the respective right hand side inner norms (see
        e.g. `CVXOPT
        <https://cvxopt.org/userguide/coneprog.html#algorithm-parameters>`__).

        To prevent that the right hand side vector norm is zero (or small),
        solvers would either add some small constant or use a fixed lower bound,
        which may be as large as :math:`1`.

        :obj:`None` lets the solver use its own default value.
        """, lambda tol: tol is None or tol > 0.0),

    ("abs_dual_fsb_tol", float, 1e-8, """
        Absolute dual problem feasibility tolerance.

        A dual solution is feasible if some norm over the vector of dual
        constraint violations is smaller than this value.

        Serves as an optimality criterion for the Simplex algorithm.

        :obj:`None` lets the solver use its own default value.
        """, lambda tol: tol is None or tol > 0.0),

    ("rel_dual_fsb_tol", float, 1e-8, """
        Relative dual problem feasibility tolerance.

        Like :ref:`abs_dual_fsb_tol <option_abs_dual_fsb_tol>`, but the norm is
        divided by the norm of the constraints' right hand side vector. (See
        :ref:`rel_prim_fsb_tol <option_rel_prim_fsb_tol>` for exceptions.)

        Serves as an optimality criterion for the Simplex algorithm.

        :obj:`None` lets the solver use its own default value.
        """, lambda tol: tol is None or tol > 0.0),

    ("abs_ipm_opt_tol", float, 1e-8, """
        Absolute optimality tolerance for interior point methods.

        Depending on the solver, a fesible primal/dual solution pair is
        considered optimal if this value upper bounds either

        - the absolute difference between the primal and dual objective values,
          or
        - the violation of the complementary slackness condition.

        The violation is computed as some norm over the vector that contains the
        products of each constraint's slack with its corresponding dual value.
        If the norm is the 1-norm, then the two conditions are equal. Otherwise
        they can differ by a factor that depends on the number and type of
        constraints.

        :obj:`None` lets the solver use its own default value.
        """, lambda tol: tol is None or tol > 0.0),

    ("rel_ipm_opt_tol", float, 1e-8, """
        Relative optimality tolerance for interior point methods.

        Like :ref:`abs_ipm_opt_tol <option_abs_ipm_opt_tol>`, but the
        suboptimality measure is divided by a convex combination of the absolute
        primal and dual objective function values.

        :obj:`None` lets the solver use its own default value.
        """, lambda tol: tol is None or tol > 0.0),

    ("abs_bnb_opt_tol", float, 1e-6, """
        Absolute optimality tolerance for branch-and-bound solution strategies
        to mixed integer problems.

        A solution is optimal if the absolute difference between the objective
        function value of the current best integer solution and the current best
        bound obtained from a continuous relaxation is smaller than this value.

        :obj:`None` lets the solver use its own default value.
        """, lambda tol: tol is None or tol > 0.0),

    ("rel_bnb_opt_tol", float, 1e-4, """
        Relative optimality tolerance for branch-and-bound solution strategies
        to mixed integer problems.

        Like :ref:`abs_bnb_opt_tol <option_abs_bnb_opt_tol>`, but the difference
        is divided by a convex combination of the absolute values of the two
        objective function values.

        :obj:`None` lets the solver use its own default value.
        """, lambda tol: tol is None or tol > 0.0),

    ("integrality_tol", float, 1e-5, r"""
        Integrality tolerance.

        A number :math:`x \in \mathbb{R}` is considered integral if
        :math:`\min_{z \in \mathbb{Z}}{|x - z|}` is at most this value.

        :obj:`None` lets the solver use its own default value.
        """, lambda tol: tol is None or (tol > 0.0 and tol < 0.5)),

    ("markowitz_tol", float, None, """
        Markowitz threshold used in the Simplex algorithm.

        :obj:`None` lets the solver use its own default value.
        """, lambda tol: tol is None or (tol > 0.0 and tol < 1.0)),

    ("max_iterations", int, None, """
        Maximum number of iterations allowed for iterative solution strategies.

        :obj:`None` means no limit.
        """, None),

    ("max_fsb_nodes", int, None, """
        Maximum number of feasible solution nodes visited for branch-and-bound
        solution strategies.

        :obj:`None` means no limit.

        .. note::

            If you want to obtain all feasible solutions that the solver
            encountered, use the :ref:`pool_size <option_pool_size>` option.
        """, None),

    ("timelimit", int, None, """
        Maximum number of seconds spent searching for a solution.

        :obj:`None` means no limit.
        """, None),

    ("lp_root_method", str, None, """
        Algorithm used to solve continuous linear problems, including the root
        relaxation of mixed integer problems.

        - :obj:`None` lets PICOS or the solver select it for you.
        - ``"psimplex"`` for Primal Simplex.
        - ``"dsimplex"`` for Dual Simplex.
        - ``"interior"`` for the interior point method.
        """, lambda value: value in (None, "psimplex", "dsimplex", "interior")),

    ("lp_node_method", str, None, """
        Algorithm used to solve continuous linear problems at non-root nodes of
        the branching tree built when solving mixed integer programs.

        - :obj:`None` lets PICOS or the solver select it for you.
        - ``"psimplex"`` for Primal Simplex.
        - ``"dsimplex"`` for Dual Simplex.
        - ``"interior"`` for the interior point method.
        """, lambda value: value in (None, "psimplex", "dsimplex", "interior")),

    ("treememory", int, None, """
        Bound on the memory used by the branch-and-bound tree, in Megabytes.

        :obj:`None` means no limit.
        """, None),

    ("pool_size", int, None, """
        Maximum number of mixed integer feasible solutions returned.

        If this is not :obj:`None`, :func:`Problem.solve <picos.Problem.solve>`
        returns a list of :class:`solutions <picos.Solution>` instead of just a
        single one.

        :obj:`None` lets the solver return only the best solution.
        """, lambda value: value is None or value >= 1),

    ("pool_rel_gap", float, None, """
        Discards solutions from the :ref:`solution pool <option_pool_size>` as
        soon as a better solution is found that beats it by the given relative
        objective function gap.

        :obj:`None` is the solver's choice, which may be *never discard*.
        """, None),

    ("pool_abs_gap", float, None, """
        Discards solutions from the :ref:`solution pool <option_pool_size>` as
        soon as a better solution is found that beats it by the given absolute
        objective function gap.

        :obj:`None` is the solver's choice, which may be *never discard*.
        """, None),

    ("hotstart", bool, False, """
        Tells the solver to start from the (partial) solution that is stored in
        the :class:`variables <.variables.BaseVariable>` assigned to the
        problem."""),

    ("verify_prediction", bool, True, """
        Whether PICOS should validate that problem reformulations produce a
        problem that matches their predicted outcome.

        If a mismatch is detected, a :class:`RuntimeError` is thrown as there is
        a chance that it is caused by a bug in the reformulation, which could
        affect the correctness of the solution. By disabling this option you are
        able to retrieve a correct solution given that the error is only in the
        prediction, and given that the solution strategy remains valid for the
        actual outcome."""),

    ("max_footprints", int, 1024, """
        Maximum number of different predicted problem formulations (footprints)
        to consider before deciding on a formulation and solver to use.

        :obj:`None` lets PICOS exhaust all reachable problem formulations.
        """, None),

    # Solver-specific options.
    # --------------------------------------------------------------------------

    ("cplex_params", dict, {}, """
        A dictionary of CPLEX parameters to be set after general options are
        passed and before the search is started.

        For example, ``{"mip.limits.cutpasses": 5}`` limits the number of
        cutting plane passes when solving the root node to :math:`5`."""),

    ("cplex_lwr_bnd_limit", float, None, """
        Tells CPLEX to stop MIP optimization if a lower bound below this value
        is found.
        """, None),

    ("cplex_upr_bnd_limit", float, None, """
        Tells CPLEX to stop MIP optimization if an upper bound above this value
        is found.
        """, None),

    ("cplex_bnd_monitor", bool, False, """
        Tells CPLEX to store information about the evolution of the bounds
        during the MIP solution search process. At the end of the computation, a
        list of triples ``(time, lowerbound, upperbound)`` will be provided in
        the field ``bounds_monitor`` of the dictionary returned by
        :func:`solve <picos.Problem.solve>`.
        """),

    ("cvxopt_kktsolver", (str, types.FunctionType), "ldl", """
        The KKT solver used by CVXOPT internally.

        See `CVXOPT's guide on exploiting structure
        <https://cvxopt.org/userguide/coneprog.html#exploiting-structure>`_.

        :obj:`None` denotes CVXOPT's default KKT solver.
        """, None),

    ("cvxopt_kktreg", float, 1e-9, """
        The KKT solver regularization term used by CVXOPT internally.

        This is an undocumented feature of CVXOPT, see `here
        <https://github.com/cvxopt/cvxopt/issues/36#issuecomment-125165634>`_.

        End of 2020, this option only affected the LDL KKT solver.

        :obj:`None` denotes CVXOPT's default value.
        """, None),

    ("gurobi_params", dict, {}, """
        A dictionary of Gurobi parameters to be set after general options are
        passed and before the search is started.

        For example, ``{"NodeLimit": 25}`` limits the number of nodes visited by
        the MIP optimizer to :math:`25`."""),

    ("mosek_params", dict, {}, """
        A dictionary of MOSEK (Optimizer) parameters to be set after general
        options are passed and before the search is started.

        See the `list of MOSEK (Optimizer) 8.1 parameters
        <https://docs.mosek.com/8.1/pythonapi/parameters.html>`_."""),

    ("mskfsn_params", dict, {}, """
        A dictionary of MOSEK (Fusion) parameters to be set after general
        options are passed and before the search is started.

        See the `list of MOSEK (Fusion) 8.1 parameters
        <https://docs.mosek.com/8.1/pythonfusion/parameters.html>`_."""),

    ("scip_params", dict, {}, """
        A dictionary of SCIP parameters to be set after general options are
        passed and before the search is started.

        For example, ``{"lp/threads": 4}`` sets the number of threads to solve
        LPs with to :math:`4`."""),
]
"""The table of available solver options.

Each entry is a tuple representing a single solver option. The tuple's entries
are, in order:

- Name of the option. Must be a valid Python attribute name.
- The option's argument type. Will be cast on any argument that is not already
  an instance of the type, except for :obj:`None`.
- The option's default value. Must already be of the proper type, or
  :obj:`None`, and must pass the optional check.
- The option's description, which is used as part of the docstring of
  :class:`Options`. In the case of a multi-line text, leading and trailing
  empty lines as well as the overall indentation are ignored.
- Optional: A boolean function used on every argument that passes the type
  conversion (so either an argument of the proper type, or :obj:`None`). If the
  function returns ``False``, then the argument is rejected. The default
  function rejects exactly :obj:`None`. Supplying :obj:`None` instead of a
  function accepts all arguments (in particular, accepts :obj:`None`).
"""

# Add per-solver options.
for name, solver in all_solvers().items():
    OPTIONS.append(("penalty_{}".format(name), float, solver.default_penalty(),
        """
        Penalty for using the {} solver.

        If solver :math:`A` has a penalty of :math:`p` and solver :math:`B` has
        a larger penality of :math:`p + x`, then :math:`B` is be chosen over
        :math:`A` only if the problem as passed to :math:`A` would be
        :math:`10^x` times larger as when passed to :math:`B`.
        """.format(name.upper())))

OPTIONS = sorted(OPTIONS)


# Explicit object inheritence is necessary for __new__ under Python 2.
class Option(object):
    """Optimization solver option.

    A single option that affects how a :class:`Problem <picos.modeling.Problem>`
    is solved.

    An initial instance of this class is built from each entry of the
    :data:`OPTIONS` table to obtain the :data:`OPTION_OBJS` tuple.
    """

    # Define __new__ in addition to __init__ so that copy can bypass __init__.
    def __new__(cls, *args, **kwargs):
        """Create a blank :class:`Option` to be filled in by :meth:`copy`."""
        return super(Option, cls).__new__(cls)

    def __init__(self, name, argType, default, description,
            check=(lambda x: x is not None)):
        """Initialize an :class:`Option`.

        See :data:`OPTIONS`.
        """
        assert default is None or isinstance(default, argType)
        assert check is None or check(default)

        self.name        = name
        self.argType     = argType
        self.default     = default
        self._value      = default
        self.description = self._normalize_description(description)
        self.check       = check

    def _normalize_description(self, description):
        lines = description.splitlines()
        notSpace = [n for n, line in enumerate(lines) if line.strip()]
        if not notSpace:
            return ""
        first, last = min(notSpace), max(notSpace)
        i = len(lines[first]) - len(lines[first].lstrip())
        return "\n".join(line[i:].rstrip() for line in lines[first:last+1])

    def _set_value(self, value):
        if value is not None and not isinstance(value, self.argType):
            if isinstance(self.argType, type):
                try:
                    value = self.argType(value)
                except Exception:
                    raise TypeError("Failed to convert argument {} to option "
                        "'{}' to type {}.".format(
                        repr(value), self.name, self.argType.__name__))
            else:
                assert isinstance(self.argType, (tuple, list))
                assert all(isinstance(t, type) for t in self.argType)
                raise TypeError("Argument {} to option '{}' does not match "
                    "permissible types {}.".format(repr(value), self.name,
                    ", ".join(t.__name__ for t in self.argType)))

        try:
            if self.check is not None and not self.check(value):
                raise ValueError()
        except Exception:
            raise ValueError("The option '{}' does not accept the value {}."
                .format(self.name, repr(value)))

        self._value = value

    value = property(lambda self: self._value, _set_value)

    def reset(self):
        """Reset the option to its default value."""
        self.value = self.default

    def is_default(self):
        """Whether the option has its default value."""
        return self.value == self.default

    def copy(self):
        """Return an independent copy of the option."""
        theCopy = self.__class__.__new__(self.__class__)
        theCopy.name        = self.name
        theCopy.argType     = self.argType
        theCopy.default     = self.default
        theCopy._value      = self._value
        theCopy.description = self.description
        theCopy.check       = self.check
        return theCopy


OPTION_OBJS = tuple(Option(*args) for args in OPTIONS)
"""The initial solver options as :class:`Option` objects."""


def _tablerow(option, indentaion=0):
    """Return a reST list-table row describing an :class:`Option`."""
    spaces = " "*indentaion
    return (
        "{}- * {{0}}\n"
        "{}  * ``{{1}}``\n"
        "{}  * .. _option_{{0}}:\n\n"
        "{}    {{2}}"
    ).format(
        *(4*(spaces,))).format(option.name, repr(option.default),
        "\n{}    ".format(spaces).join(option.description.splitlines()))


def _jumplabel(option):
    """Return a reStructuredText jumplabel describing an :class:`Option`."""
    return ":ref:`{0} <option_{0}>`".format(option.name)


# Explicit object inheritence is necessary for __new__ under Python 2.
class Options(object):
    """Collection of optimization solver options.

    A collection of options that affect how a :class:`Problem
    <picos.modeling.Problem>` is solved. :attr:`Problem.options
    <picos.Problem.options>` is an instance of this class.

    The options can be accessed as an attribute or as an item. The latter
    approach supports Unix shell-style wildcard characters:

    >>> import picos
    >>> P = picos.Problem()
    >>> P.options.verbosity = 2
    >>> P.options["primals"] = False
    >>> # Set all absolute tolerances at once.
    >>> P.options["abs_*_tol"] = 1e-6

    There are two corresponding ways to reset an option to its default value:

    >>> del P.options.verbosity
    >>> P.options.reset("primals", "*_tol")

    Options can also be passed as a keyword argument sequence when the
    :class:`Problem <picos.Problem>` is created and whenever a solution is
    searched:

    >>> # Use default options except for verbosity.
    >>> P = picos.Problem(verbosity = 1)
    >>> x = P.add_variable("x", lower = 0); P.set_objective("min", x)
    >>> # Only for the next search: Don't be verbose anyway.
    >>> solution = P.solve(solver = "cvxopt", verbosity = 0)
    """

    # Document the individual options.
    __doc__ += \
    """
    .. rubric:: Available Options

    Jump to option: ➥\xa0{}

    .. list-table::
      :header-rows: 1
      :widths: 10 10 80

      - * Option
        * Default
        * Description
    """.format(" ➥\xa0".join(_jumplabel(option) for option in OPTION_OBJS)) \
    .rstrip() + "\n" + "\n".join(_tablerow(option, 6) for option in OPTION_OBJS)

    # Define __new__ in addition to __init__ so that
    # 1. __init__ does not take the static default options as an argument,
    #    hiding them from the user and the documentation while
    # 2. Options.copy can still bypass copying the default options (by bypassing
    #    __init__) so that options aren't copied twice.
    def __new__(cls, *args, **kwargs):
        """Create an empty options set."""
        instance = super(Options, cls).__new__(cls)
        # Options overwrites __setattr__, so we need to call object.__setattr__.
        super(Options, cls).__setattr__(instance, "_options", {})
        return instance

    def __init__(self, **options):
        """Create a default option set and set the given options on top."""
        for option in OPTION_OBJS:
            self._options[option.name] = option.copy()

        self.update(**options)

    def __str__(self):
        defaults = sorted(
            (o for o in self._options.values() if o.is_default()),
            key=(lambda o: o.name))
        modified = sorted(
            (o for o in self._options.values() if not o.is_default()),
            key=(lambda o: o.name))

        nameLength  = max(len(o.name)       for o in self._options.values())
        valueLength = max(len(str(o.value)) for o in self._options.values())

        string = ""

        if modified:
            defaultLength = max(len(str(o.default)) for o in modified)

            string += "Modified solver options:\n" + "\n".join((
                "  {{:{}}} = {{:{}}} (default: {{:{}}})".format(
                    nameLength, valueLength, defaultLength
                ).format(
                    option.name, str(option.value), str(option.default))
                for num, option in enumerate(modified)))

        if defaults:
            if modified:
                string += "\n\n"

            string += "Default solver options:\n" + "\n".join((
                "  {{:{}}} = {{}}".format(nameLength).format(
                    option.name, str(option.value))
                for num, option in enumerate(defaults)))

        return string

    def __eq__(self, other):
        if self is other:
            return True

        for name in self._options:
            if self._options[name].value != other._options[name].value:
                return False

        return True

    def _fuzzy(returnsSomething):
        """Allow wildcards in option names."""
        def decorator(method):
            def wrapper(self, pattern, *extraArgs):
                if any(char in pattern for char in "*?[!]"):
                    matching = fnmatch.filter(self._options.keys(), pattern)

                    if not matching:
                        raise LookupError("No option matches '{}'."
                            .format(pattern))

                    if returnsSomething:
                        return {name: method(self, name, *extraArgs)
                            for name in matching}
                    else:
                        for name in matching:
                            method(self, name, *extraArgs)
                else:
                    if returnsSomething:
                        return method(self, pattern, *extraArgs)
                    else:
                        method(self, pattern, *extraArgs)
            return wrapper
        return decorator

    @_fuzzy(True)
    def __getattr__(self, name):
        if name in self._options:
            return self._options[name].value
        else:
            raise AttributeError("Unknown option '{}'.".format(name))

    @_fuzzy(False)
    def __setattr__(self, name, value):
        if name in self._options:
            self._options[name].value = value
        else:
            raise AttributeError("Unknown option '{}'.".format(name))

    @_fuzzy(False)
    def __delattr__(self, name):
        if name in self._options:
            self._options[name].reset()
        else:
            raise AttributeError("Unknown option '{}'.".format(name))

    @_fuzzy(True)
    def __getitem__(self, name):
        if name in self._options:
            return self._options[name].value
        else:
            raise LookupError("Unknown option '{}'.".format(name))

    @_fuzzy(False)
    def __setitem__(self, name, value):
        if name in self._options:
            self._options[name].value = value
        else:
            raise LookupError("Unknown option '{}'.".format(name))

    def __contains__(self, name):
        return name in self._options

    def __dir__(self):
        optionNames = [name for name in self._options.keys()]
        if hasattr(super(Options, self), "__dir__"):  # Python 3
            list = super(Options, self).__dir__() + optionNames
        else:  # Python 2
            list = dir(self.__class__) + self.__dict__.keys() + optionNames
        return sorted(list)

    def copy(self):
        """Return an independent copy of the current options set."""
        theCopy = self.__class__.__new__(self.__class__)
        for option in self._options.values():
            theCopy._options[option.name] = option.copy()
        return theCopy

    def update(self, **options):
        """Set multiple options at once.

        This method is called with the keyword arguments supplied to the
        :class:`Options` constructor, so the following two are the same:

        >>> import picos
        >>> a = picos.Options(verbosity = 1, primals = False)
        >>> b = picos.Options()
        >>> b.update(verbosity = 1, primals = False)
        >>> a == b
        True

        :param options: A parameter sequence of options to set.
        """
        for key, val in options.items():
            self[key] = val

    def updated(self, **options):
        """Return a modified copy."""
        theCopy = self.copy()
        if options:
            theCopy.update(**options)
        return theCopy

    def self_or_updated(self, **options):
        """Return either a modified copy or self, depending on given options."""
        if options:
            theCopy = self.copy()
            theCopy.update(**options)
            return theCopy
        else:
            return self

    @_fuzzy(False)
    def _reset_single(self, name):
        self._options[name].reset()

    def reset(self, *options):
        """Reset all or a selection of options to their default values.

        :param options: The names of the options to reset, may contain wildcard
            characters. If no name is given, all options are reset.
        """
        if options:
            for name in options:
                self._reset_single(name)
        else:
            for option in self._options.values():
                option.reset()

    @_fuzzy(True)
    def _help_single(self, name):
        option = self._options[name]
        return (
            "Option:  {}\n"
            "Default: {}\n"
            "\n  {}"
        ).format(option.name, str(option.default),
            "\n  ".join(option.description.splitlines()))

    def help(self, *options):
        """Print text describing selected options.

        :param options: The names of the options to describe, may contain
            wildcard characters.
        """
        for i, name in enumerate(options):
            if i != 0:
                print("\n\n")
            retval = self._help_single(name)
            if isinstance(retval, str):
                print(retval)
            else:
                assert isinstance(retval, dict)
                print("\n\n".join(retval.values()))

    @property
    def nondefaults(self):
        """A dictionary mapping option names to nondefault values.

        :Example:

        >>> from picos import Options
        >>> o = Options()
        >>> o.verbosity = 2
        >>> o.nondefaults
        {'verbosity': 2}
        >>> Options(**o.nondefaults) == o
        True
        """
        return {name: option._value for name, option in self._options.items()
            if option._value != option.default}


# --------------------------------------
__all__ = api_end(_API_START, globals())



# coding: utf-8

# ------------------------------------------------------------------------------
# Copyright (C) 2018-2019 Maximilian Stahlberg
#
# This file is part of PICOS.
#
# PICOS is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PICOS is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# ------------------------------------------------------------------------------

"""Optimization constraint types.

This package contains the constraint types that are used to express optimization
constraints. You do not need to instanciate these constraints directly; it is
more convenient to create them by applying Python's comparison operators to
algebraic expressions (see the :ref:`cheatsheet`).
"""

from ..apidoc import api_end, api_start


_API_START = api_start(globals())
# -------------------------------


# Import the base classes.
from .constraint import ConstraintType, Constraint, ConstraintConversion  # noqa

# Import the implementation classes.
from .con_absolute  import AbsoluteValueConstraint  # noqa
from .con_affine    import ComplexAffineConstraint, AffineConstraint  # noqa
from .con_detrootn  import DetRootNConstraint  # noqa
from .con_expcone   import ExpConeConstraint  # noqa
from .con_flow      import FlowConstraint  # noqa
from .con_geomean   import GeometricMeanConstraint  # noqa
from .con_kldiv     import KullbackLeiblerConstraint  # noqa
from .con_lmi       import ComplexLMIConstraint, LMIConstraint  # noqa
from .con_log       import LogConstraint  # noqa
from .con_logsumexp import LogSumExpConstraint  # noqa
from .con_matnorm   import (MatrixNormConstraint, # noqa
                            SpectralNormConstraint, NuclearNormConstraint)
from .con_powtrace  import PowerTraceConstraint  # noqa
from .con_quadratic import (NonconvexQuadraticConstraint, # noqa
                           ConvexQuadraticConstraint, ConicQuadraticConstraint)
from .con_rsoc      import RSOCConstraint  # noqa
from .con_simplex   import SimplexConstraint  # noqa
from .con_soc       import SOCConstraint  # noqa
from .con_sumexp    import SumExponentialsConstraint  # noqa
from .con_sumxtr    import SumExtremesConstraint  # noqa
from .con_vecnorm   import VectorNormConstraint  # noqa


# --------------------------------------
__all__ = api_end(_API_START, globals())



# coding: utf-8

# ------------------------------------------------------------------------------
# Copyright (C) 2019 Maximilian Stahlberg
#
# This file is part of PICOS.
#
# PICOS is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PICOS is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# ------------------------------------------------------------------------------

"""Functions to quickly solve a problem."""

from . import Problem
from ..apidoc import api_end, api_start

_API_START = api_start(globals())
# -------------------------------


def _make_doc(minimize):
    """Create the docstring for :func:`minimize` and :func:`maximize`."""
    lower = "minimize" if minimize else "maximize"
    capital = lower.capitalize()
    sign = "" if minimize else "-"

    return \
    """{capital} a scalar expression subject to constraints.

    Internally, this creates a :class:`~picos.Problem`,
    :meth:`sets an objective <picos.Problem.set_objective>`,
    :meth:`adds constraints <picos.Problem.add_list_of_constraints>`, performs a
    :meth:`solution search <picos.Problem.solve>` and returns an optimum value
    found.

    :param ~picos.expressions.Expression function:
        The objective function to {lower}.

    :param list(~picos.constraints.Constraint) subject_to:
        A collection of constraints to obey.

    :param options:
        A keyword argument sequence of solver options to use. See
        :class:`~picos.Options`.

    :returns:
        The optimum value, as computed from an applied solution.

    :raises ~picos.SolutionFailure:
        See :meth:`~picos.Problem.solve`.

    :Example:

    >>> from picos import {lower}, RealVariable
    >>> x = RealVariable("x")
    >>> p = {lower}({sign}x**2, [(x - 2)**2 <= x - 2], solver="cvxopt")
    >>> round(p, 5)
    {sign}4.0
    >>> round(x, 5)
    2.0
    """.format(capital=capital, lower=lower, sign=sign)


def minimize(function, subject_to = [], **options):  # noqa
    P = Problem(**options)
    P.set_objective("min", function)
    P.add_list_of_constraints(subject_to)
    P.solve()
    return P.value


def maximize(function, subject_to = [], **options):  # noqa
    P = Problem(**options)
    P.set_objective("max", function)
    P.add_list_of_constraints(subject_to)
    P.solve()
    return P.value


minimize.__doc__ = _make_doc(minimize=True)
maximize.__doc__ = _make_doc(minimize=False)


def find_assignment(subject_to=[], **options):
    """Find a feasible variable assignment.

    Internally, this creates a :class:`~picos.Problem`,
    :meth:`adds constraints <picos.Problem.add_list_of_constraints>` and
    performs a :meth:`solution search <picos.Problem.solve>`.

    :param list(~picos.constraints.Constraint) subject_to:
        A collection of constraints to obey.

    :param options:
        A keyword argument sequence of solver options to use. See
        :class:`~picos.Options`.

    :returns:
        Nothing. Check the concerned variables'
        :attr:`values <picos.expressions.Expression.value>`.

    :raises ~picos.SolutionFailure:
        See :meth:`~picos.Problem.solve`.

    :Example:

    >>> from picos import find_assignment, RealVariable
    >>> x = RealVariable("x")
    >>> values = find_assignment([x**2 + 1 <= x], solver="cvxopt")
    ... # doctest: +ELLIPSIS
    Traceback (most recent call last):
        ...
    picos.modeling.problem.SolutionFailure: Code 3: ...
    >>> x.value
    >>> find_assignment([x**2 + 0.25 <= x], solver="cvxopt")
    >>> round(x.value, 5)
    0.5
    """
    P = Problem(**options)
    P.add_list_of_constraints(subject_to)
    P.solve()


# --------------------------------------
__all__ = api_end(_API_START, globals())



# coding: utf-8

# ------------------------------------------------------------------------------
# Copyright (C) 2019 Maximilian Stahlberg
# Based on the original picos.expressions module by Guillaume Sagnol.
#
# This file is part of PICOS.
#
# PICOS is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PICOS is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# ------------------------------------------------------------------------------

"""Implements :class:`PowerTrace`."""

import operator
from collections import namedtuple

import cvxopt
import numpy

from .. import glyphs
from ..apidoc import api_end, api_start
from ..caching import cached_unary_operator
from ..constraints import Constraint, PowerTraceConstraint
from .data import (convert_and_refine_arguments, convert_operands, cvx2np,
                   cvxopt_hpsd, make_fraction)
from .exp_affine import AffineExpression, ComplexAffineExpression, Constant
from .expression import Expression, refine_operands, validate_prediction

_API_START = api_start(globals())
# -------------------------------


class PowerTrace(Expression):
    r"""The trace of the :math:`p`-th power of a hermitian matrix.

    :Definition:

    Let :math:`p \in \mathbb{Q}`.

    1.  If the base expressions is a real scalar :math:`x` and no additional
        constant :math:`m` is given, then this is the power :math:`x^p`.

    2.  If the base expressions is a real scalar :math:`x`,
        :math:`p \in [0, 1]`, and a positive scalar constant :math:`m` is given,
        then this is the scaled power :math:`m x^p`.

    3.  If the base expression is a hermitian matrix :math:`X` and no additional
        constant :math:`M` is given, then this is the trace of power
        :math:`\operatorname{tr}(X^p)`.

    4.  If the base expression is a hermitian matrix :math:`X`,
        :math:`p \in [0, 1]`, and a hermitian positive semidefinite constant
        matrix :math:`M` of same shape as :math:`X` is given, then this is the
        trace of a scaled power :math:`\operatorname{tr}(M X^p)`.

    No other case is supported. In particular, if :math:`p \not\in [0, 1]`, then
    :math:`m`/:math:`M` must be undefined (:obj:`None`).

    .. warning::

        1. For a constraint of the form :math:`x^p \leq t` with :math:`p < 1`
           and :math:`p \neq 0`, PICOS enforces :math:`x \geq 0` during solution
           search.

        2. For a constraint of the form :math:`\operatorname{tr}(X^p) \leq t` or
           :math:`\operatorname{tr}(M X^p) \leq t` with :math:`p < 1` and
           :math:`p \neq 0`, PICOS enforces :math:`X \succeq 0` during solution
           search.

        3. For a constraint of the form :math:`\operatorname{tr}(X^p) \leq t`
           or :math:`\operatorname{tr}(M X^p) \leq t` with :math:`p > 1`, PICOS
           enforces :math:`t \geq 0` during solution search.
    """

    # --------------------------------------------------------------------------
    # Initialization and factory methods.
    # --------------------------------------------------------------------------

    @convert_and_refine_arguments("x")
    def __init__(self, x, p, m=None, denominator_limit=1000):
        """Construct a :class:`PowerTrace`.

        :param x: The scalar or symmetric matrix to form a power of.
        :type x: ~picos.expressions.AffineExpression
        :param float p: The value for :math:`p`, which is cast to a limited
            precision fraction.
        :param m: An additional positive semidefinite constant to multiply the
            power with.
        :type m: :class:`~picos.expressions.AffineExpression` or anything
            recognized by :func:`~picos.expressions.data.load_data`
        :param int denominator_limit: The largest allowed denominator when
            casting :math:`p` to a fraction. Higher values can yield a greater
            precision at reduced performance.
        """
        # Validate x.
        if not isinstance(x, ComplexAffineExpression):
            raise TypeError("Can only form the power of an affine expression, "
                "not of {}.".format(x.string))
        elif not x.square:
            raise TypeError(
                "Can't form the power of non-square {}.".format(x.string))
        elif not x.hermitian:
            raise NotImplementedError("Taking {} to a power is not supported "
                "as it is not necessarily hermitian.".format(x.string))

        # Load p.
        pNum, pDen, p, pStr = make_fraction(p, denominator_limit)

        # Load m.
        if m is not None:
            mStr = "m" if len(x) == 1 else "M"

            if p < 0 or p > 1:
                raise ValueError(
                    "p-th power with an additional factor {} requires {}."
                    .format(mStr, glyphs.le(0, glyphs.le("p", 1))))

            if not isinstance(m, ComplexAffineExpression):
                try:
                    m = Constant(mStr, m, x.shape)
                except Exception as error:
                    raise TypeError("Failed to load the additional factor {} "
                        "as a matrix of same shape as {}: {}"
                        .format(mStr, x.string, error))
            else:
                m = m.refined

            if not m.constant:
                raise TypeError("The additional factor {} is not constant."
                    .format(m.string))
            elif not cvxopt_hpsd(m.value):
                raise ValueError("The additional factor {} is not hermitian "
                    "positive semidefinite.".format(m.string))

        self._x     = x
        self._num   = pNum
        self._den   = pDen
        self._m     = m
        self._limit = denominator_limit

        if m is None:
            if len(x) == 1:
                typeStr = "Power"
                if p == 2:
                    symbStr = glyphs.squared(x.string)
                elif p == 3:
                    symbStr = glyphs.cubed(x.string)
                else:
                    symbStr = glyphs.power(x.string, pStr)
            else:
                typeStr = "Trace of Power"
                symbStr = glyphs.trace(glyphs.power(x.string, pStr))
        else:
            if len(x) == 1:
                typeStr = "Scaled Power"
                symbStr = glyphs.mul(m.string, glyphs.power(x.string, pStr))
            else:
                typeStr = "Trace of Scaled Power"
                symbStr = glyphs.trace(glyphs.mul(
                    m.string, glyphs.power(x.string, pStr)))

        Expression.__init__(self, typeStr, symbStr)

    # --------------------------------------------------------------------------
    # Abstract method implementations and method overridings, except _predict.
    # --------------------------------------------------------------------------

    def _get_refined(self):
        if self._x.constant:
            return Constant(self._symbStr, self.value)
        elif self.p == 0:
            if self._m is not None:
                return self._m.tr
            else:
                return Constant(
                    glyphs.Fn("diaglen({})")(self._x.string), self._x.shape[0])
        elif self.p == 1:
            if self._m is not None:
                # NOTE: No hermitian transpose as both m and x are hermitian.
                return (self._m | self._x)
            else:
                return self._x.tr
        elif self.p == 2 and self._x.scalar and self._m is None:
            return (self._x | self._x)
        else:
            return self

    Subtype = namedtuple("Subtype", ("diag", "num", "den", "hasM", "complex"))

    def _get_subtype(self):
        return self.Subtype(
            self._x.shape[0], self._num, self._den, self._m is not None,
            self._x.complex)

    def _get_value(self):
        x = cvx2np(self._x._get_value())
        p = self.p

        eigenvalues = numpy.linalg.eigvalsh(x)

        if p != int(p) and any(value < 0 for value in eigenvalues):
            raise ArithmeticError("Cannot evaluate {}: {} is not positive "
                "semidefinite and the exponent is fractional."
                .format(self.string, self._x.string))

        if self._m is None:
            trace = sum([value**p for value in eigenvalues])
        else:
            m = cvx2np(self._m._get_value())

            U, S, V = numpy.linalg.svd(x)
            power = U*numpy.diag(S**p)*V
            trace = numpy.trace(m * power)

        return cvxopt.matrix(trace)

    @cached_unary_operator
    def _get_variables(self):
        return self._x.variables

    def _is_convex(self):
        return self.p >= 1 or self.p <= 0

    def _is_concave(self):
        return self.p >= 0 and self.p <= 1

    def _replace_variables(self, var_map):
        return self.__class__(
            self._x._replace_variables(var_map), self.p, self._m, self._limit)

    # --------------------------------------------------------------------------
    # Python special method implementations, except constraint-creating ones.
    # --------------------------------------------------------------------------

    @convert_operands(scalarRHS=True)
    @refine_operands()
    def __mul__(self, other):
        if isinstance(other, AffineExpression):
            if not other.constant:
                raise NotImplementedError("You may only multiply a nonconstant "
                    "PICOS trace of a power with a constant term.")

            value = other.value

            if value < 0:
                raise NotImplementedError("You may only multiply a nonconstant "
                    "PICOS trace of a power with a nonnegative term.")

            if value == 1:
                return self

            if self._m is None:
                m = other.dupdiag(self.n).renamed(other.string)
            else:
                m = other*self._m

            return PowerTrace(self._x, self.p, m, self._limit)
        else:
            return NotImplemented

    @convert_operands(scalarRHS=True)
    @refine_operands()
    def __rmul__(self, other):
        if isinstance(other, AffineExpression):
            return self.__mul__(other)
        else:
            return NotImplemented

    # --------------------------------------------------------------------------
    # Methods and properties that return expressions.
    # --------------------------------------------------------------------------

    @property
    def x(self):
        """The matrix concerned."""
        return self._x

    # --------------------------------------------------------------------------
    # Methods and properties that describe the expression.
    # --------------------------------------------------------------------------

    @property
    def n(self):
        """Diagonal length of :attr:`x`."""
        return self._x.shape[0]

    @property
    def p(self):
        """The parameter :math:`p`.

        This is a limited precision version of the paramter used when the
        expression was constructed.
        """
        return float(self._num) / float(self._den)

    @property
    def num(self):
        """The limited precision fraction numerator of :math:`p`."""
        return self._num

    @property
    def den(self):
        """The limited precision fraction denominator of :math:`p`."""
        return self._den

    @property
    def m(self):
        """An additional factor to multiply the power with."""
        return self._m

    # --------------------------------------------------------------------------
    # Constraint-creating operators, and _predict.
    # --------------------------------------------------------------------------

    @classmethod
    def _predict(cls, subtype, relation, other):
        assert isinstance(subtype, cls.Subtype)

        p = float(subtype.num) / float(subtype.den)

        if relation == operator.__le__:
            if p > 0 and p < 1:  # Not convex.
                return NotImplemented

            if issubclass(other.clstype, AffineExpression) \
            and other.subtype.dim == 1:
                return PowerTraceConstraint.make_type(*subtype)
        elif relation == operator.__ge__:
            if p < 0 or p > 1:  # Not concave.
                return NotImplemented

            if issubclass(other.clstype, AffineExpression) \
            and other.subtype.dim == 1:
                return PowerTraceConstraint.make_type(*subtype)

        return NotImplemented

    @convert_operands(scalarRHS=True)
    @validate_prediction
    @refine_operands()
    def __le__(self, other):
        if not self.convex:
            raise TypeError("Cannot upper-bound a nonconvex (trace of) power.")

        if isinstance(other, AffineExpression):
            return PowerTraceConstraint(self, Constraint.LE, other)
        else:
            return NotImplemented

    @convert_operands(scalarRHS=True)
    @validate_prediction
    @refine_operands()
    def __ge__(self, other):
        if not self.concave:
            raise TypeError("Cannot lower-bound a nonconcave (trace of) power.")

        if isinstance(other, AffineExpression):
            return PowerTraceConstraint(self, Constraint.GE, other)
        else:
            return NotImplemented


# --------------------------------------
__all__ = api_end(_API_START, globals())



# coding: utf-8

# ------------------------------------------------------------------------------
# Copyright (C) 2019 Maximilian Stahlberg
# Based on the original picos.expressions module by Guillaume Sagnol.
#
# This file is part of PICOS.
#
# PICOS is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PICOS is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# ------------------------------------------------------------------------------

"""Implements :class:`Ball`."""

import operator
from collections import namedtuple

from .. import glyphs
from ..apidoc import api_end, api_start
from .data import convert_and_refine_arguments, convert_operands, make_fraction
from .exp_affine import AffineExpression, ComplexAffineExpression, Constant
from .exp_norm import Norm
from .expression import refine_operands, validate_prediction
from .set import Set

_API_START = api_start(globals())
# -------------------------------


class Ball(Set):
    r"""A ball of radius :math:`r` according to a (generalized) :math:`p`-norm.

    :Definition:

    In the following, :math:`\lVert \cdot \rVert_p` refers to the vector
    :math:`p`-norm or to the entrywise matrix :math:`p`-norm, depending on the
    argument. See :class:`~picos.Norm` for definitions.

    Let :math:`r \in \mathbb{R}`.

    1.  For :math:`p \in [1, \inf)`, this is the convex set

        .. math::

            \{x \in \mathbb{K} \mid \lVert x \rVert_p \leq r\}

        for any

        .. math::

            \mathbb{K} \in \bigcup_{m, n \in \mathbb{Z}_{\geq 1}}
                \left( \mathbb{C}^n \cup \mathbb{C}^{m \times n} \right).

    2.  For a generalized :math:`p`-norm with :math:`p \in (0, 1)`, this is the
        convex set

        .. math::

            \{x \in \mathbb{R} \mid \lVert x \rVert_p \geq r \land x \geq 0\}

        for any

        .. math::

            \mathbb{K} \in \bigcup_{m, n \in \mathbb{Z}_{\geq 1}}
                \left( \mathbb{R}^n \cup \mathbb{R}^{m \times n} \right).

    Note that :math:`x` may not be complex if :math:`p < 1` due to the implicit
    :math:`x \geq 0` constraint in this case, which is not meaningful on the
    complex field.

    Note further that :math:`r` may be any scalar affine expression, it does not
    need to be constant.
    """

    @convert_and_refine_arguments("radius")
    def __init__(self, radius=Constant(1), p=2, denominator_limit=1000):
        """Construct a :math:`p`-norm ball of given radius.

        :param radius: The ball's radius.
        :type radius:
            float or ~picos.expressions.AffineExpression
        :param float p: The value for :math:`p`, which is cast to a limited
            precision fraction.
        :param int denominator_limit: The largest allowed denominator when
            casting :math:`p` to a fraction. Higher values can yield a greater
            precision at reduced performance.
        """
        num, den, p, pStr = make_fraction(p, denominator_limit)

        if not isinstance(radius, AffineExpression):
            raise TypeError("The ball's radius must be given as a real affine "
                "expression, not as {}.".format(type(radius).__name__))
        elif not radius.scalar:
            raise TypeError("The ball's radius must be scalar, not of shape {}."
                .format(glyphs.shape(radius.shape)))

        var  = glyphs.free_var_name(radius.string)
        unit = "Unit " if radius.is1 else ""
        if p >= 1:
            typeStr = "{}{}-norm Ball" \
                .format(unit, pStr if den == 1 else glyphs.parenth(pStr))
            symbStr = glyphs.set(glyphs.sep(
                var, glyphs.le(glyphs.pnorm(var, pStr), radius.string)))
        else:
            typeStr = "Nonneg. Compl. of {}{}-norm Ball" \
                .format(unit, pStr if den == 1 else glyphs.parenth(pStr))
            symbStr = glyphs.set(glyphs.sep(glyphs.ge(var, 0),
                glyphs.ge(glyphs.pnorm(var, pStr), radius.string)))

        self._num    = num
        self._den    = den
        self._limit  = denominator_limit
        self._radius = radius

        Set.__init__(self, typeStr, symbStr)

    @property
    def p(self):
        """The value :math:`p` defining the :math:`p`-norm used.

        This is a limited precision version of the paramter used when the ball
        was constructed.
        """
        return float(self._num) / float(self._den)

    @property
    def r(self):
        """The ball's radius :math:`r`."""
        return self._radius

    def _get_variables(self):
        return self._radius.variables

    def _replace_variables(self, var_map):
        return self.__class__(
            self.p, self._radius._replace_variables(var_map), self._limit)

    Subtype = namedtuple("Subtype", ("num", "den"))

    def _get_subtype(self):
        return self.Subtype(self._num, self._den)

    @classmethod
    def _predict(cls, subtype, relation, other):
        assert isinstance(subtype, cls.Subtype)

        num = subtype.num
        den = subtype.den
        p   = float(num) / float(den)

        if relation == operator.__rshift__:
            if issubclass(other.clstype, ComplexAffineExpression):
                complex = not issubclass(other.clstype, AffineExpression)

                if complex and p < 1:
                    return NotImplemented

                # The shape of the real, vectorized version of the element.
                shape = (other.subtype.dim * (2 if complex else 1), 1)

                norm = Norm.make_type(shape, num, den, num, den)

                # HACK: Whether the radius is constant makes no difference.
                radius = AffineExpression.make_type((1, 1), None, None)

                if p >= 1:
                    return norm <= radius
                else:
                    return norm >= radius

        return NotImplemented

    @convert_operands()
    @validate_prediction
    @refine_operands()
    def __rshift__(self, element):
        if isinstance(element, ComplexAffineExpression):
            if element.complex and self.p < 1:
                raise TypeError("Cannot constrain a complex expression to be "
                    "in the nonnegative complement of a generalized p-norm "
                    "ball: Nonnegativity is not clear.")

            norm = Norm(element, self.p, denominator_limit=self._limit)

            if self.p >= 1:
                return norm <= self._radius
            else:
                return norm >= self._radius
        else:
            return NotImplemented


# --------------------------------------
__all__ = api_end(_API_START, globals())



# coding: utf-8

# ------------------------------------------------------------------------------
# Copyright (C) 2019 Maximilian Stahlberg
#
# This file is part of PICOS.
#
# PICOS is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PICOS is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# ------------------------------------------------------------------------------

"""Backwards-compatibility and deprecation helpers."""

import functools
import warnings

from .apidoc import api_end, api_start
from .formatting import doc_cat

_API_START = api_start(globals())
# -------------------------------


def throw_deprecation_warning(reason, decoratorLevel=0):
    """Raise a deprecation warning."""
    warnings.warn(reason, DeprecationWarning, stacklevel=3 + decoratorLevel)


def deprecated(since, reason=None, useInstead=None, decoratorLevel=0):
    """Decorate a deprecated function or method.

    :param str reason: Textual deprecation reason.
    :param str useInstead: Alternative to the deprecated object.
    :param int decoratorLevel: If not the only decorator, number of decorators
        above this one that add to the stack level of a warning thrown.
    """
    assert len([x for x in (reason, useInstead) if x is not None]) == 1, \
        "Need to supply exactly one reason or reference to @deprecated."

    if reason is not None:
        rest, text = reason, reason
    if useInstead is not None:
        rest = "Use :py:obj:`{0}` instead.".format(useInstead)
        text = "Use {} instead.".format(useInstead.split(".")[-1]
            if useInstead.startswith("~") else useInstead)

    def decorator(wrapped):
        if hasattr(wrapped, "__qualname__"):  # Python 3
            name = wrapped.__qualname__
        else:  # Python 2
            name = wrapped.__name__

        # Make the wrapper copy the wrapped object's documentation …
        @functools.wraps(wrapped)
        def wrapper(*args, **kwargs):
            warnings.warn(
                "{} is deprecated: {}".format(name, text),
                DeprecationWarning, stacklevel=2 + decoratorLevel)
            return wrapped(*args, **kwargs)

        # … and extend it.
        wrapper.__doc__ = doc_cat(
            wrapper.__doc__ if wrapper.__doc__ else "Deprecated name.",
            ".. deprecated:: {}\n    {}".format(since, rest))

        return wrapper
    return decorator


_option_name_old2new = {
    "allow_license_warnings": "license_warnings",
    "verbose":      "verbosity",
    "noprimals":    "primals",
    "noduals":      "duals",
    "tol":          ("*_fsb_tol", "*_ipm_opt_tol"),
    "gaplim":       "rel_bnb_opt_tol",
    "maxit":        "max_iterations",
    "nbsol":        "max_fsb_nodes",
    "pool_relgap":  "pool_rel_gap",
    "pool_absgap":  "pool_abs_gap",
    "lboundlimit":  "cplex_lwr_bnd_limit",
    "uboundlimit":  "cplex_upr_bnd_limit",
    "boundMonitor": "cplex_bnd_monitor",
    "solve_via_dual": "dualize"
}

_option_value_old2new = {
    "noprimals": lambda x: not x,
    "noduals": lambda x: not x,
    "solve_via_dual": lambda x: bool(x)  # Map None to False.
}


def map_legacy_options(options={}, **kwOptions):
    """Map deprecated solver options to those replacing them."""
    kwOptions.update(options)

    newOptions = {}
    for name, value in kwOptions.items():
        if name in _option_name_old2new:
            newNames = _option_name_old2new[name]

            if isinstance(newNames, str):
                newNames = (newNames,)

            warnings.warn(
                "Option '{}' is deprecated, use '{}'.".format(
                name, ", ".join(newNames)), DeprecationWarning, stacklevel=3)

            if name in _option_value_old2new:
                value = _option_value_old2new[name](value)

            for newName in newNames:
                newOptions[newName] = value
        else:
            newOptions[name] = value

    return newOptions


# --------------------------------------
__all__ = api_end(_API_START, globals())



# coding: utf-8

# ------------------------------------------------------------------------------
# Copyright (C) 2019 Maximilian Stahlberg
# Based on the original picos.expressions module by Guillaume Sagnol.
#
# This file is part of PICOS.
#
# PICOS is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PICOS is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# ------------------------------------------------------------------------------

"""Implements :class:`DetRootN`."""

import operator
from collections import namedtuple

import cvxopt
import numpy

from .. import glyphs
from ..apidoc import api_end, api_start
from ..caching import cached_unary_operator
from ..constraints import DetRootNConstraint
from .data import convert_and_refine_arguments, convert_operands, cvx2np
from .exp_affine import AffineExpression, ComplexAffineExpression
from .expression import Expression, refine_operands, validate_prediction

_API_START = api_start(globals())
# -------------------------------


class DetRootN(Expression):
    r"""The :math:`n`-th root of the determinant of an :math:`n\times n` matrix.

    :Definition:

    For an :math:`n \times n` positive semidefinite hermitian matrix :math:`X`,
    this is

    .. math::

        \sqrt[n]{\det X}.

    .. warning::

        When you pose a lower bound on the :math:`n`-th root of a determinant of
        the matrix :math:`X`, then PICOS enforces positive semidefiniteness
        :math:`X \succeq 0` through an auxiliary constraint during solution
        search.
    """

    # --------------------------------------------------------------------------
    # Initialization and factory methods.
    # --------------------------------------------------------------------------

    @convert_and_refine_arguments("x")
    def __init__(self, x):
        """Construct a :class:`DetRootN`.

        :param x: The matrix concerned. Must be hermitian by definition.
        :type x: ~picos.expressions.ComplexAffineExpression
        """
        if not isinstance(x, ComplexAffineExpression):
            raise TypeError("Can only form the determinant of an affine "
                "expression, not of {}.".format(type(x).__name__))
        elif not x.square:
            raise TypeError("Can't take the determinant of non-square {0}."
                .format(x.string))
        elif not x.hermitian:
            raise NotImplementedError("Taking the n-th root of the determinant "
                "of {0} is not supported as {0} is not necessarily hermitian."
                .format(x.string))

        self._x = x

        Expression.__init__(self, "n-th Root of a Determinant",
            glyphs.power(glyphs.det(x.string), glyphs.div(1, x.shape[0])))

    # --------------------------------------------------------------------------
    # Abstract method implementations and method overridings, except _predict.
    # --------------------------------------------------------------------------

    def _get_refined(self):
        if self._x.constant:
            return AffineExpression.from_constant(self.value, 1, self._symbStr)
        elif len(self._x) == 1:
            return self._x.renamed(self._symbStr)
        else:
            return self

    Subtype = namedtuple("Subtype", ("diag", "complex"))

    def _get_subtype(self):
        return self.Subtype(self.n, self._x.complex)

    def _get_value(self):
        value = self._x._get_value()

        det = numpy.linalg.det(cvx2np(value))

        if det < 0:
            raise ArithmeticError("Cannot evaluate {}: {} is negative."
                .format(self.string, glyphs.eq(glyphs.det(self.x.string), det)))

        return cvxopt.matrix(det**(1.0 / self._x.shape[0]))

    @cached_unary_operator
    def _get_variables(self):
        return self._x.variables

    def _is_convex(self):
        return False

    def _is_concave(self):
        return True

    def _replace_variables(self, var_map):
        return self.__class__(self._x._replace_variables(var_map))

    # --------------------------------------------------------------------------
    # Python special method implementations, except constraint-creating ones.
    # --------------------------------------------------------------------------

    @convert_operands(scalarRHS=True)
    @refine_operands()
    def __mul__(self, other):
        if isinstance(other, AffineExpression):
            if not other.constant:
                raise NotImplementedError("You may only multiply a nonconstant "
                    "PICOS n-th root of a determinant with a constant term.")

            root = DetRootN(other.value * self._x)
            root._typeStr = "Scaled " + root._typeStr
            root._symbStr = glyphs.clever_mul(self.string, other.string)
            return root
        else:
            return NotImplemented

    @convert_operands(scalarRHS=True)
    @refine_operands()
    def __rmul__(self, other):
        if isinstance(other, AffineExpression):
            mean = self.__mul__(other)
            # NOTE: __mul__ always creates a fresh expression.
            mean._symbStr = glyphs.clever_mul(other.string, self.string)
            return mean
        else:
            return NotImplemented

    # --------------------------------------------------------------------------
    # Methods and properties that return modified copies.
    # --------------------------------------------------------------------------

    @property
    def x(self):
        """The matrix concerned."""
        return self._x

    # --------------------------------------------------------------------------
    # Methods and properties that describe the expression.
    # --------------------------------------------------------------------------

    @property
    def n(self):
        """Diagonal length of :attr:`x`."""
        return self._x.shape[0]

    # --------------------------------------------------------------------------
    # Constraint-creating operators, and _predict.
    # --------------------------------------------------------------------------

    @classmethod
    def _predict(cls, subtype, relation, other):
        assert isinstance(subtype, cls.Subtype)

        if relation == operator.__ge__:
            if issubclass(other.clstype, AffineExpression) \
            and other.subtype.dim == 1:
                return DetRootNConstraint.make_type(
                    diag=subtype.diag, complex=subtype.complex)

        return NotImplemented

    @convert_operands(scalarRHS=True)
    @validate_prediction
    @refine_operands()
    def __ge__(self, other):
        if isinstance(other, AffineExpression):
            return DetRootNConstraint(self, other)
        else:
            return NotImplemented


# --------------------------------------
__all__ = api_end(_API_START, globals())



# coding: utf-8

# ------------------------------------------------------------------------------
# Copyright (C) 2019 Maximilian Stahlberg
# Based on the original picos.expressions module by Guillaume Sagnol.
#
# This file is part of PICOS.
#
# PICOS is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PICOS is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# ------------------------------------------------------------------------------

"""Implements :class:`QuadraticExpression`."""

import operator
import sys
from collections import OrderedDict as odict
from collections import namedtuple

import cvxopt
import cvxopt.cholmod
import numpy

from .. import glyphs
from ..apidoc import api_end, api_start
from ..caching import (borrow_cache, cached_property,
                       cached_selfinverse_unary_operator,
                       cached_unary_operator)
from ..constraints import (ConicQuadraticConstraint, Constraint,
                           ConvexQuadraticConstraint,
                           NonconvexQuadraticConstraint)
from .data import convert_operands, cvx2np, should_be_sparse
from .exp_affine import AffineExpression, ComplexAffineExpression, Constant
from .expression import Expression, refine_operands, validate_prediction
from .variables import BaseVariable

_API_START = api_start(globals())
# -------------------------------


PSD_PERTURBATIONS = 3
r"""Maximum number of singular quadratic form perturbations.

PICOS uses NumPy either or CHOLMOD to compute a Cholesky decomposition of a
positive semidefinite quadratic form :math:`Q`. Both libraries require
:math:`Q` to be nonsingular, which is not a requirement on PICOS' end.
If either library rejects :math:`Q`, then PICOS provides a sequence of
:data:`PSD_PERTURBATIONS` perturbed quadratic forms :math:`Q + \epsilon I` for
increasing :math:`\epsilon` until the perturbed matrix is found positive
definite or until the largest :math:`\epsilon` was tested unsuccessfully.

If this is zero, PICOS will only decompose quadratic forms that are nonsingular.
If this is one, then only the largest epsilon is tested.
"""


class QuadraticExpression(Expression):
    """A scalar quadratic expression of the form :math:`x^TQx + a^Tx + b`."""

    # --------------------------------------------------------------------------
    # Initialization and factory methods.
    # --------------------------------------------------------------------------

    def __init__(
            self, string, quadraticPart={}, affinePart=AffineExpression.zero(),
            scalarFactors=None, copyDecomposition=None):
        r"""Initialize a scalar quadratic expression.

        This constructor is meant for internal use. As a user, you will most
        likely want to build expressions starting from
        :mod:`~picos.expressions.variables` or a :func:`~picos.Constant`.

        :param str string: A symbolic string description.
        :param quadraticPart: A :class:`dict` mapping PICOS variable pairs to
            CVXOPT matrices. Each entry :math:`(x, y) \mapsto A_{xy}` represents
            a quadratic form
            :math:`\operatorname{vec_x}(x)^T A_{xy} \operatorname{vec_y}(y)`
            where :math:`\operatorname{vec_x}` and :math:`\operatorname{vec_y}`
            refer to the isometric real
            :mod:`~.picos.expressions.vectorizations` that are used by PICOS
            internally to store the variables :math:`x` and :math:`y`,
            respectively. The quadratic part :math:`Q` of the expression is then
            given as :math:`Q = \sum_{x, y} A_{xy}`.
        :param affinePart: The affine part :math:`a^Tx + b` of the expression.
        :type affinePart: ~picos.expressions.AffineExpression
        :param scalarFactors: A pair :math:`(u, v)` with both :math:`u` and
            :math:`v` scalar real affine expressions representing a known
            :math:`x^TQx + a^Tx + b = uv` factorization of the expression.
        :type factorization:
            tuple(~picos.expressions.AffineExpression)
        :param copyDecomposition: Another quadratic expression with equal
            quadratic part whose quadratic part decomposition shall be copied.
        :type copyDecomposition:
            ~picos.expressions.QuadraticExpression
        """
        # Ensure correct usage within PICOS.
        assert isinstance(quadraticPart, dict) and all(
            isinstance(xy, (tuple, list)) and len(xy) == 2
            and isinstance(xy[0], BaseVariable)
            and isinstance(xy[1], BaseVariable)
            and isinstance(A, (cvxopt.matrix, cvxopt.spmatrix))
            and A.size == (xy[0].dim, xy[1].dim)
            for xy, A in quadraticPart.items())
        assert isinstance(affinePart, ComplexAffineExpression) \
            and len(affinePart) == 1
        assert not copyDecomposition \
            or isinstance(copyDecomposition, QuadraticExpression)
        assert not scalarFactors or (
            isinstance(scalarFactors, (tuple, list)) and len(scalarFactors) == 2
            and all(isinstance(x, AffineExpression) for x in scalarFactors)
            and all(len(x) == 1 for x in scalarFactors))

        Expression.__init__(self, "Quadratic Expression", string)

        # Check affine part.
        affinePart = affinePart.refined
        if not isinstance(affinePart, AffineExpression):
            raise NotImplementedError(
                "PICOS does not support complex-valued quadratic expressions "
                "and the affine part of {} is not real.".format(self._symbStr))
        elif len(affinePart) != 1:
            raise TypeError("Affine part of {} is not scalar."
                .format(self._symbStr))

        # Store quadratic part.
        self._quads = {}
        for xy, A in quadraticPart.items():
            # Ignore coefficients that are already zero.
            if not A:
                continue

            if xy[0] is xy[1]:
                # Store the unique symmetric form for two reasons:
                # - Hermitian forms concerning the original variables are
                #   supplied as quadratic forms concerning the isometric real
                #   vectorizations of the variables. These matrices are in
                #   general still complex, but their symmetric form is not.
                # - Coefficients that cancel out are set to numeric zeros and
                #   can be converted to structural zeros by _sparse_quads.
                A = 0.5*(A + A.T)
            if hash(xy[1]) < hash(xy[0]):
                # Store only one coefficient per unordered pair.
                xy, A = (xy[1], xy[0]), A.T

            if xy in self._quads:
                self._quads[xy] = self._quads[xy] + A
            else:
                self._quads[xy] = A

        # Remove coefficients of zero.
        self._quads = {xy: A for xy, A in self._quads.items() if A}

        # Check if coefficients are real.
        for (x, y), A in self._quads.items():
            if A.typecode == "z":
                if any(A.imag()):
                    raise NotImplementedError(
                        "PICOS does not support complex-valued quadratic "
                        "expressions and the quadratic part of {} for variables"
                        " {} and {} is not real or equivalent to a real form."
                        .format(self._symbStr, x.string, y.string))

                self._quads[xy] = A.real()

        # Store affine part.
        self._aff = affinePart

        # Store a known factorization into real scalars.
        if scalarFactors:
            a, b = scalarFactors
            if a.equals(b):
                self._sf = (a, a)  # Speedup future equality checks.
            else:
                self._sf = (a, b)
        else:
            self._sf = None

        # Copy a decomposition from another quadratic expression.
        if copyDecomposition is not None:
            borrow_cache(self, copyDecomposition, ("_Q_and_x", "L", "quadroot"))

    # --------------------------------------------------------------------------
    # Abstract method implementations and method overridings, except _predict.
    # --------------------------------------------------------------------------

    @cached_unary_operator
    def _get_refined(self):
        if not self._quads:
            return self._aff.refined.renamed(self._symbStr)
        else:
            return self

    Subtype = namedtuple("Subtype", (
        "convex", "concave", "qterms", "rank", "haslin", "factors"))

    @cached_unary_operator
    def _get_subtype(self):
        convex  = self.convex
        concave = False if convex and self._quads else self.concave
        qterms  = self.num_quad_terms
        haslin  = not self._aff.constant

        try:
            if convex:
                rank = self.rank
            elif concave:
                rank = (-self).rank
            else:
                rank = None
        except ValueError:  # No quadratic part.
            rank = 0

        factors = len(set(self.scalar_factors))

        return self.Subtype(convex, concave, qterms, rank, haslin, factors)

    def _get_value(self):
        value = self._aff._get_value()

        for xy, A in self._quads.items():
            x, y = xy
            xVal = x._get_internal_value()
            yVal = y._get_internal_value()
            value = value + xVal.T * A * yVal

        return value

    @cached_unary_operator
    def _get_variables(self):
        return self._aff.variables.union(
            var for vars in self._quads for var in vars)

    def _is_convex(self):
        try:
            self.L
        except ValueError:
            return True  # Expression is affine.
        except ArithmeticError:
            return False
        else:
            return True

    def _is_concave(self):
        try:
            (-self).L
        except ValueError:
            return True  # Expression is affine.
        except ArithmeticError:
            return False
        else:
            return True

    def _replace_variables(self, var_map):
        name_map = {old.name: new.name for old, new in var_map.items()}

        string = self.string
        if isinstance(string, glyphs.GlStr):
            string = string.reglyphed(name_map)

        quads = {(var_map[var1], var_map[var2]): coef
            for (var1, var2), coef in self._quads.items()}

        if not self._sf:
            sf = None
        elif self._sf[0] is self._sf[1]:
            sf = (self._sf[0]._replace_variables(var_map),)*2
        else:
            sf = tuple(f._replace_variables(var_map) for f in self._sf)

        return QuadraticExpression(
            string, quads, self.aff._replace_variables(var_map), sf)

    # --------------------------------------------------------------------------
    # Python special method implementations, except constraint-creating ones.
    # --------------------------------------------------------------------------

    def __len__(self):
        # Faster version that overrides Expression.__len__.
        return 1

    @convert_operands(sameShape=True)
    def __add__(self, other):
        """Denote addition from the right hand side."""
        if isinstance(other, QuadraticExpression):
            string = glyphs.clever_add(self.string, other.string)

            quadPart = {}
            varPairs = set(self._quads.keys()).union(other._quads.keys())
            for vars in varPairs:
                if vars in self._quads:
                    if vars in other._quads:
                        quadPart[vars] = self._quads[vars] + other._quads[vars]
                    else:
                        quadPart[vars] = self._quads[vars]
                else:
                    quadPart[vars] = other._quads[vars]

            affPart = self._aff + other._aff

            return QuadraticExpression(string, quadPart, affPart)
        elif isinstance(other, AffineExpression):
            string = glyphs.clever_add(self.string, other.string)

            return QuadraticExpression(
                string, self._quads, self._aff + other, copyDecomposition=self)
        else:
            return NotImplemented

    @convert_operands(sameShape=True)
    def __radd__(self, other):
        """Denote addition from the left hand side."""
        if isinstance(other, QuadraticExpression):
            assert False, "Impossible case."
        elif isinstance(other, AffineExpression):
            result = self.__add__(other)
            # NOTE: __add__ always creates a fresh expression.
            result._symbStr = glyphs.clever_add(other.string, self.string)
            return result
        else:
            return NotImplemented

    @convert_operands(sameShape=True)
    def __sub__(self, other):
        """Denote substraction from the right hand side."""
        if isinstance(other, QuadraticExpression):
            string = glyphs.clever_sub(self.string, other.string)

            quadPart = {}
            varPairs = set(self._quads.keys()).union(other._quads.keys())
            for vars in varPairs:
                if vars in self._quads:
                    if vars in other._quads:
                        quadPart[vars] = self._quads[vars] - other._quads[vars]
                    else:
                        quadPart[vars] = self._quads[vars]
                else:
                    quadPart[vars] = -other._quads[vars]

            affPart = self._aff - other._aff

            return QuadraticExpression(string, quadPart, affPart)
        elif isinstance(other, AffineExpression):
            string = glyphs.clever_sub(self.string, other.string)

            return QuadraticExpression(
                string, self._quads, self._aff - other, copyDecomposition=self)
        else:
            return NotImplemented

    @convert_operands(sameShape=True)
    def __rsub__(self, other):
        """Denote substraction with self on the right hand side."""
        if isinstance(other, QuadraticExpression):
            assert False, "Impossible case."
        elif isinstance(other, AffineExpression):
            result = (-self).__add__(other)
            # NOTE: __add__ always creates a fresh expression.
            result._symbStr = glyphs.clever_sub(other.string, self.string)
            return result
        else:
            return NotImplemented

    @cached_selfinverse_unary_operator
    def __neg__(self):
        string   = glyphs.clever_neg(self.string)
        quadPart = {xy: -A for xy, A in self._quads.items()}
        affPart  = -self._aff
        factors  = (self._sf[0], -self._sf[1]) if self._sf else None

        return QuadraticExpression(string, quadPart, affPart, factors)

    @convert_operands(scalarRHS=True)
    def __mul__(self, other):
        """Denote scaling from the right hand side."""
        if isinstance(other, AffineExpression):
            if not other.constant:
                raise NotImplementedError("You may only multiply a PICOS "
                    "quadratic expression with a constant term.")

            factor   = other.value
            string   = glyphs.clever_mul(self.string, other.string)
            quadPart = {xy: factor*A for xy, A in self._quads.items()}
            affPart  = self._aff*other

            if self._sf:
                r = factor**0.5
                if self._sf[0] is self._sf[1]:
                    factors = (r*self._sf[0],)*2
                else:
                    factors = (r*self._sf[0], r*self._sf[1])
            else:
                factors = None

            return QuadraticExpression(string, quadPart, affPart, factors)
        else:
            return NotImplemented

    @convert_operands(scalarRHS=True)
    def __rmul__(self, other):
        """Denote scaling from the left hand side."""
        if isinstance(other, AffineExpression):
            result = self.__mul__(other)
            # NOTE: __mul__ always creates a fresh expression.
            result._symbStr = glyphs.clever_mul(other.string, self.string)
            return result
        else:
            return NotImplemented

    @convert_operands(scalarRHS=True)
    def __truediv__(self, other):
        """Denote division by a constant scalar."""
        if isinstance(other, AffineExpression):
            if not other.constant:
                raise NotImplementedError("You may only divide a PICOS "
                    "quadratic expression by a constant term.")

            result = self.__mul__(1.0 / other.value)
            # NOTE: __mul__ always creates a fresh expression.
            result._symbStr = glyphs.div(self.string, other.string)
            return result
        else:
            return NotImplemented

    def __div__(self, other):
        """Denote division by a scalar constant."""
        # Python 2 uses this instead of __truediv__.
        return self.__truediv__(other)

    # --------------------------------------------------------------------------
    # Decomposition related methods.
    # --------------------------------------------------------------------------

    @cached_property
    def _sparse_quads(self):
        """Quadratic coefficients (re-)cast to sparse matrices."""
        return {xy: cvxopt.sparse(M) for xy, M in self._quads.items()}

    def _Q_and_x_or_A_and_y(self, augmentedMatrix):
        """Either :attr:`_Q_and_x` or :attr:`_A_and_y`."""
        # Find occurences of the scalar entries of the variable's isometric real
        # vectorization in the quadratic part.
        nonzero = {}
        for (x, y), A in self._sparse_quads.items():
            nonzero.setdefault(x, set())
            nonzero.setdefault(y, set())

            nonzero[x].update(A.I)
            nonzero[y].update(A.J)

        # For the augmented matrix, find additional linear occurences.
        if augmentedMatrix:
            for x, a in self._aff._sparse_coefs.items():
                nonzero.setdefault(x, set())
                nonzero[x].update(a.I)

        nonzero = odict((var, sorted(nonzero[var]))
            for var in sorted(nonzero, key=lambda v: v.id))

        # Compute each (multidimensional) variable's offset in Q/A.
        offset, offsets = 0, {}
        for var, nz in nonzero.items():
            offsets[var] = offset
            offset += len(nz)

        # Compute Q.
        V, I, J = [], [], []
        for (x, y), A in self._sparse_quads.items():
            B = A[nonzero[x], nonzero[y]]

            V.extend(B.V)
            I.extend(offsets[x] + i for i in B.I)
            J.extend(offsets[y] + j for j in B.J)

        # Turn Q into A if requested.
        if augmentedMatrix:
            for x, a in self._aff._sparse_coefs.items():
                b = a[nonzero[x]]

                V.extend(b.V)
                I.extend(offsets[x] + i for i in b.I)
                J.extend([offset]*len(b.J))

            V.append(self._aff._const[0])
            I.append(offset)
            J.append(offset)

            offset += 1

        # Finalize Q/A.
        Q = cvxopt.spmatrix(V, I, J, (offset,)*2, tc="d")
        Q = 0.5*(Q + Q.T)

        if not Q:
            if augmentedMatrix:
                raise ValueError(
                    "The expression {} is zero.".format(self._symbStr))
            else:
                raise ValueError(
                    "The quadratic part of {} is zero.".format(self._symbStr))

        # Compute x.
        x = None
        for var, nz in nonzero.items():
            n, r = len(nz), var.dim
            F = cvxopt.spmatrix([1]*n, range(n), nz, (n, r))
            y = AffineExpression(
                glyphs.Fn("F")(var.string), n, coefficients={var: F})
            x = y if x is None else x // y

        # Turn x into y if requested.
        if augmentedMatrix:
            x = AffineExpression.from_constant(1) if x is None else x // 1

        return Q, x

    @cached_property
    def _Q_and_x(self):
        """Pair containing :attr:`Q` and :attr:`x`."""
        return self._Q_and_x_or_A_and_y(augmentedMatrix=False)

    @cached_property
    def _A_and_y(self):
        """Pair containing :attr:`R` and :attr:`y`."""
        return self._Q_and_x_or_A_and_y(augmentedMatrix=True)

    @property
    def Q(self):
        """The coefficient matrix :math:`Q` of the expression, condensed.

        This equals the :math:`Q` of the quadratic expression
        :math:`x^TQx + a^Tx + b` but with zero rows and columns removed.

        The vector :attr:`x` is a condensed version of :math:`x` that refers to
        this matrix, so that ``q.x.T*q.Q*q.x`` equals the quadratic part
        :math:`x^TQx` of the expression ``q``.

        :raises ValueError: When the quadratic part is zero.
        """
        return self._Q_and_x[0]

    @property
    def x(self):
        """The stacked variable vector :math:`x` of the expression, condensed.

        This equals the :math:`x` of the quadratic expression
        :math:`x^TQx + a^Tx + b` but entries corresponding to zero rows and
        columns in :attr:`Q` are removed, so that ``q.x.T*q.Q*q.x`` equals the
        :math:`x^TQx` part of the expression ``q``.

        :raises ValueError: When the quadratic part is zero.
        """
        return self._Q_and_x[1]

    @property
    def A(self):
        r"""An affine-augmented quadratic coefficient matrix, condensed.

        For a quadratic expression :math:`x^TQx + a^Tx + b`, this is
        :math:`A = \begin{bmatrix}Q&\frac{a}{2}\\\frac{a^T}{2}&b\end{bmatrix}`
        but with zero rows and columns removed.

        The vector :attr:`y` is a condensed version of :math:`x` that refers to
        this matrix, so that ``q.y.T*q.A*q.y`` equals the expression ``q``.

        :raises ValueError: When the expression is zero.
        """
        return self._A_and_y[0]

    @property
    def y(self):
        """See :attr:`A`.

        :raises ValueError: When the expression is zero.
        """
        return self._A_and_y[1]

    def _L_or_M(self, augmentedMatrix):
        """Either :attr:`L` or :attr:`M`."""
        Q = self.A if augmentedMatrix else self.Q
        n = Q.size[0]

        # Define a sequence of small numbers for perturbation.
        # TODO: These numbers are empirical; find worst case values.
        spread   = [abs(v) for v in Q.V if v]
        minEps   = 1 * min(spread) * sys.float_info.epsilon
        maxEps   = n * max(spread) * sys.float_info.epsilon
        epsilons = set([0.0])
        for i in range(PSD_PERTURBATIONS):
            epsilons.add(minEps + i * (maxEps - minEps) / PSD_PERTURBATIONS)
        epsilons = sorted(epsilons)

        # Check if Q is really sparse.
        sparse = should_be_sparse(Q.size, len(Q))

        # Attempt to find L.
        L = None
        if sparse:  # Use CHOLMOD.
            F = None
            for eps in epsilons:
                P = Q + cvxopt.spdiag([eps]*n) if eps else Q  # Perturbed Q.

                if F is None:
                    F = cvxopt.cholmod.symbolic(P)

                try:
                    cvxopt.cholmod.numeric(P, F)
                except ArithmeticError:
                    pass
                else:
                    L = cvxopt.cholmod.getfactor(F)
                    break
        else:  # Use NumPy.
            Q = cvx2np(Q)
            for eps in epsilons:
                P = Q + numpy.diag([eps]*n) if eps else Q  # Perturbed Q.

                try:
                    L = numpy.linalg.cholesky(P)
                except numpy.linalg.LinAlgError:
                    pass
                else:
                    break

        if L is None:
            if not PSD_PERTURBATIONS:
                raise ArithmeticError("The expression {} has an (augmented) "
                    "quadratic form that is either not positive semidefinite "
                    "or singular. Try enabling PSD_PERTURBATIONS.")
            elif augmentedMatrix:
                raise ArithmeticError("The expression {} is not a squared norm "
                    "as its {} representation is not positive semidefinite."
                    .format(self._symbStr, glyphs.matrix(glyphs.vertcat(
                        glyphs.horicat("Q", glyphs.div("a", 2)),
                        glyphs.horicat(glyphs.div(glyphs.transp("a"), 2), "b")
                    ))))
            else:
                raise ArithmeticError("The expression {} is not convex as its "
                    "quadratic part is not positive semidefinite."
                    .format(self._symbStr))

        # Remove near-zeros in the order of n·sqrt(maxEps).
        cutoff = 2 * n * maxEps**0.5
        if sparse:
            VIJ = list(t for t in zip(L.V, L.I, L.J) if abs(t[0]) > cutoff)
            if VIJ:  # Don't zero the matrix.
                V, I, J = zip(*VIJ)
                L = cvxopt.spmatrix(V, I, J, L.size)
        else:
            mask = abs(L) <= cutoff
            if not numpy.all(mask):  # Don't zero the matrix.
                L[mask] = 0
            L = cvxopt.sparse(cvxopt.matrix(L))

        return L

    @cached_property
    def L(self):
        """The :math:`L` of an :math:`LL^T` Cholesky decomposition of :attr:`Q`.

        :returns: A CVXOPT lower triangular sparse matrix.

        :raises ValueError: When the quadratic part is zero.
        :raises ArithmeticError: When the expression is not convex, that is when
            the matrix :attr:`Q` is not numerically positive semidefinite.
        """
        return self._L_or_M(augmentedMatrix=False)

    @cached_property
    def M(self):
        """The :math:`M` of an :math:`MM^T` Cholesky decomposition of :attr:`A`.

        :returns: A CVXOPT lower triangular sparse matrix.

        :raises ValueError: When the expression is zero.
        :raises ArithmeticError: When the expression can not be written as a
            squared norm, that is when the extended matrix :attr:`A` is not
            numerically positive semidefinite.
        """
        return self._L_or_M(augmentedMatrix=True)

    @cached_property
    def quadroot(self):
        r"""Affine expression whose squared norm equals :math:`x^TQx`.

        For a convex quadratic expression ``q``, this is equal to the vector
        ``q.L.T*q.x`` with zero rows removed.

        :Construction:

        Let :math:`x^TQx` be the quadratic part of the expression with :math:`Q`
        positive semidefinite and :math:`Q = LL^T` a Cholesky decomposition.
        Then,

        .. math::
            x^TQx
            &= x^TLL^Tx \\
            &= (L^Tx)^TL^Tx \\
            &= \langle L^Tx, L^Tx \rangle \\
            &= \lVert L^Tx \rVert^2.

        Note that removing zero rows from :math:`L^Tx` does not affect the norm.

        :raises ValueError: When the quadratic part is zero.
        :raises ArithmeticError: When the expression is not convex, that is when
            the quadratic part is not numerically positive semidefinite.
        """
        L   = self.L
        n   = L.size[0]
        nz  = sorted(set(L.J))
        nnz = len(nz)

        # F removes rows corresponding to zero columns in L.
        F = cvxopt.spmatrix([1.0]*nnz, range(nnz), nz, (nnz, n))

        result = (F*L.T)*self.x
        # NOTE: AffineExpression.__mul__ always creates a fresh expression.
        result._symbStr = glyphs.Fn("quadroot({})")(self.string)
        return result

    @cached_property
    def fullroot(self):
        r"""Affine expression whose squared norm equals the expression.

        For a convex quadratic expression ``q``, this is equal to the vector
        ``q.M.T*q.y`` with zero rows removed.

        :Construction:

        For a quadratic expression :math:`x^TQx + a^Tx + b` with
        :math:`A = \begin{bmatrix}Q&\frac{a}{2}\\\frac{a^T}{2}&b\end{bmatrix}`
        positive semidefinite, let :math:`A = MM^T` be a Cholesky decomposition.
        Let further :math:`y = \begin{bmatrix}x^T & 1\end{bmatrix}^T`. Then,

        .. math::
            x^TQx + a^Tx + b
            &= y^TAy \\
            &= y^TMM^Ty \\
            &= (M^Ty)^TM^Ty \\
            &= \langle M^Ty, M^Ty \rangle \\
            &= \lVert M^Ty \rVert^2.

        Note that removing zero rows from :math:`M^Ty` does not affect the norm.

        :raises ValueError: When the expression is zero.
        :raises ArithmeticError: When the expression is not convex, that is when
            the quadratic part is not numerically positive semidefinite.
        """
        M   = self.M
        n   = M.size[0]
        nz  = sorted(set(M.J))
        nnz = len(nz)

        # F removes rows corresponding to zero columns in M.
        F = cvxopt.spmatrix([1.0]*nnz, range(nnz), nz, (nnz, n))

        result = (F*M.T)*self.y
        # NOTE: AffineExpression.__mul__ always creates a fresh expression.
        result._symbStr = glyphs.Fn("fullroot({})")(self.string)
        return result

    @property
    def rank(self):
        """The length of the vector :attr:`quadroot`.

        Up to numerical considerations, this is the rank of the (convex)
        quadratic coefficient matrix :math:`Q` of the expression.

        :raises ArithmeticError: When the expression is not convex, that is when
            the quadratic part is not numerically positive semidefinite.
        """
        try:
            return len(self.quadroot)
        except ValueError:
            return 0  # No quadratic part.

    @cached_property
    def num_quad_terms(self):
        """The number of terms in the simplified quadratic form."""
        num_terms = 0
        for A in self._sparse_quads.values():
            for i, j in zip(A.I, A.J):
                if i >= j:
                    num_terms += 1
        return num_terms

    @property
    def is_squared_norm(self):
        r"""Whether the expression can be written as a squared norm.

        If this is :obj:`True`, then the there is a coefficient vector
        :math:`c` such that

        .. math::
            \lVert c^T \begin{bmatrix} x \\ 1 \end{bmatrix} \rVert^2
            = x^TQx + a^Tx + b.

        If the expression is also nonzero, then :attr:`fullroot` is
        :math:`c^T \begin{bmatrix} x^T & 1 \end{bmatrix}^T` with zero entries
        removed.
        """
        try:
            self.M
        except ValueError:
            assert self.is0
            return True
        except ArithmeticError:
            return False
        else:
            return True

    # --------------------------------------------------------------------------
    # Properties and functions that describe the expression.
    # --------------------------------------------------------------------------

    @property
    def quadratic_forms(self):
        """The quadratic forms as sparse matrices."""
        return dict(self._sparse_quads)

    @property
    def is0(self):
        """Whether the quadratic expression is zero."""
        return not self._quads and self._aff.is0

    @property
    def scalar_factors(self):
        r"""Decomposition into scalar real affine expressions.

        If the expression is known to be equal to :math:`ab` for scalar real
        affine expressions :math:`a` and :math:`b`, returns the pair ``(a, b)``.
        Otherwise, returns the empty tuple ``()``.

        Note that if :math:`a = b`, then also ``a is b`` in the returned tuple.
        """
        if self._sf:
            return tuple(self._sf)
        else:
            return ()

    # --------------------------------------------------------------------------
    # Methods and properties that return modified copies.
    # --------------------------------------------------------------------------

    @cached_property
    def quad(self):
        """Quadratic part :math:`x^TQx` of the quadratic expression."""
        factors = self._sf if self._sf and self._aff.is0 else None

        return QuadraticExpression(glyphs.quadpart(self._symbStr), self._quads,
            scalarFactors=factors, copyDecomposition=self)

    @cached_property
    def aff(self):
        """Affine part :math:`a^Tx + b` of the quadratic expression."""
        return self._aff.renamed(glyphs.affpart(self._symbStr))

    @cached_property
    def lin(self):
        """Linear part :math:`a^Tx` of the quadratic expression."""
        return AffineExpression(glyphs.linpart(self._symbStr),
            coefficients=self._aff._coefs)

    @cached_property
    def cst(self):
        """Constant part :math:`b` of the quadratic expression."""
        return AffineExpression(glyphs.cstpart(self._symbStr),
            constant=self._aff._const)

    # --------------------------------------------------------------------------
    # Constraint-creating operators, and _predict.
    # --------------------------------------------------------------------------

    @classmethod
    def _predict(cls, subtype, relation, other):
        assert isinstance(subtype, cls.Subtype)

        if issubclass(other.clstype, QuadraticExpression):
            raise NotImplementedError("Constraint outcome prediction not "
                "supported when comparing two quadratic expressions.")

        if relation == operator.__le__:
            if issubclass(other.clstype, AffineExpression) \
            and other.subtype.dim == 1:
                if subtype.convex:
                    return ConvexQuadraticConstraint.make_type(
                        qterms=subtype.qterms, rank=subtype.rank,
                        haslin=(subtype.haslin or not other.subtype.constant))
                else:
                    return NonconvexQuadraticConstraint.make_type(
                        qterms=subtype.qterms)
        elif relation == operator.__ge__:
            if issubclass(other.clstype, AffineExpression) \
            and other.subtype.dim == 1:
                if subtype.concave:
                    return ConvexQuadraticConstraint.make_type(
                        qterms=subtype.qterms, rank=subtype.rank,
                        haslin=(subtype.haslin or not other.subtype.constant))
                elif subtype.factors and other.subtype.constant \
                and other.subtype.nonneg:
                    # See __ge__ for the variants below.

                    # Variant 1:
                    return ConicQuadraticConstraint.make_type(
                        qterms=subtype.qterms, conic_argdim=1,
                        rotated=(subtype.factors == 2))

                    # Variant 2:
                    # return RSOCConstraint.make_type(argdim=1)

                    # Variant 3:
                    # if subtype.factors == 1:
                    #     return AffineConstraint.make_type(dim=1, eq=False)
                    # else:
                    #     return RSOCConstraint.make_type(argdim=1)
                else:
                    return NonconvexQuadraticConstraint.make_type(
                        qterms=subtype.qterms)

        return NotImplemented

    @convert_operands(sameShape=True)
    @validate_prediction
    @refine_operands()
    def __le__(self, other):
        LE = Constraint.LE

        if isinstance(other, QuadraticExpression):
            le0 = self - other  # Unpredictable outcome.
            if le0.convex:
                return ConvexQuadraticConstraint(self, LE, other)
            elif self.is_squared_norm and other.scalar_factors:
                return ConicQuadraticConstraint(self, LE, other)
            else:
                return NonconvexQuadraticConstraint(self, LE, other)
        elif isinstance(other, AffineExpression):
            if self.convex:
                return ConvexQuadraticConstraint(self, LE, other)
            else:
                return NonconvexQuadraticConstraint(self, LE, other)
        else:
            return NotImplemented

    @convert_operands(sameShape=True)
    @validate_prediction
    @refine_operands()
    def __ge__(self, other):
        GE = Constraint.GE

        if isinstance(other, QuadraticExpression):
            le0 = other - self  # Unpredictable outcome.
            if le0.convex:
                return ConvexQuadraticConstraint(self, GE, other)
            elif other.is_squared_norm and self.scalar_factors:
                return ConicQuadraticConstraint(self, GE, other)
            else:
                return NonconvexQuadraticConstraint(self, GE, other)
        elif isinstance(other, AffineExpression):
            if self.concave:
                return ConvexQuadraticConstraint(self, GE, other)
            elif self._sf and other.constant and other.value >= 0:
                # Handle the case of c <= x*x and c <= x*y with c pos. constant.
                root = Constant(glyphs.sqrt(other.string), other.value**0.5)

                # Variant 1: Return a ConicQuadraticConstraint that can be
                #     reformulated to an (R)SOCConstraint. This is consistent
                #     with all other quadratic expression comparison outcomes.
                other = QuadraticExpression(other.string, affinePart=other,
                    scalarFactors=(root, root))  # Unrefined QE.
                return ConicQuadraticConstraint(self, GE, other)

                # Variant 2: Return an RSOCConstraint. This is somewhat
                #     consistent with Norm returning an SOCConstraint and it is
                #     also legacy behavior. However this may not be what the
                #     user wants, because it silently forces x, y >= 0!
                # return RSOCConstraint(root, self._sf[0], self._sf[1],
                #     customString = glyphs.le(other.string, self.string))

                # Variant 3: Return either an RSOCConstraint or an affine
                #     constraint that equals an SOCConstraint. This is faster
                #     than Variant 2 but even more confusing, because the
                #     constraint would transform from e.g. 1 <= x*x to
                #     sqrt(1) <= x without any mention of cones.
                # if self._sf[0] is self._sf[1]:
                #     # NOTE: The following is not allowed by SOCConstraint
                #     #       since this is effectively an affine constraint.
                #     # return SOCConstraint(root, self._sf[0])
                #
                #     return root <= self._sf[0]
                # else:
                #     return RSOCConstraint(root, *self._sf)
            else:
                return NonconvexQuadraticConstraint(self, GE, other)
        else:
            return NotImplemented


# --------------------------------------
__all__ = api_end(_API_START, globals())



# coding: utf-8

# ------------------------------------------------------------------------------
# Copyright (C) 2019 Maximilian Stahlberg
#
# This file is part of PICOS.
#
# PICOS is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PICOS is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# ------------------------------------------------------------------------------

"""Functions to support the automatic API documentation.

The environment variable ``PICOS_WARN_MISSING_DOCSTRINGS`` can be set to warn
about missing docstrings for top level objects of a module.
"""

import os
import warnings
from types import ModuleType


_API_START = set(globals())  # api_start is not yet defined.
# -------------------------


def api_start(theGlobals):
    """Mark the start of a module's content.

    Store the returned value and pass it to :func:`api_end`.
    """
    return set(theGlobals)


def api_end(startGlobals, endGlobals):
    """Mark the end of a module's content.

    Store the returned value in __all__.
    """
    names = sorted(set(name for name, obj in endGlobals.items()
        if not name.startswith("_") and not isinstance(obj, ModuleType))
        .difference(startGlobals))

    if "PICOS_WARN_MISSING_DOCSTRINGS" in os.environ:
        for name in names:
            if not endGlobals[name].__doc__:
                warnings.warn(
                    "Top level object {}.{} has empty docstring."
                    .format(endGlobals["__name__"], name), stacklevel=3)

    return names


# --------------------------------------
__all__ = api_end(_API_START, globals())



# coding: utf-8

# ------------------------------------------------------------------------------
# Copyright (C) 2018-2019 Maximilian Stahlberg
#
# This file is part of PICOS.
#
# PICOS is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PICOS is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# ------------------------------------------------------------------------------

"""Implementation of :class:`LogSumExpConstraint`."""

from collections import namedtuple

from .. import glyphs
from ..apidoc import api_end, api_start
from ..caching import cached_property
from .constraint import Constraint, ConstraintConversion

_API_START = api_start(globals())
# -------------------------------


class LogSumExpConstraint(Constraint):
    """Upper bound on a logarithm of a sum of exponentials."""

    class ExpConeConversion(ConstraintConversion):
        """Bound on a log-sum-exp to exponential cone constraint conversion."""

        @classmethod
        def predict(cls, subtype, options):
            """Implement :meth:`~.constraint.ConstraintConversion.predict`."""
            from ..expressions import RealVariable
            from . import AffineConstraint, ExpConeConstraint

            n = subtype.argdim

            yield ("var", RealVariable.make_var_type(dim=n, bnd=0), 1)
            yield ("con", AffineConstraint.make_type(dim=1, eq=False), 1)
            yield ("con", ExpConeConstraint.make_type(), n)

        @classmethod
        def convert(cls, con, options):
            """Implement :meth:`~.constraint.ConstraintConversion.convert`."""
            from ..expressions import ExponentialCone
            from ..modeling import Problem

            x = con.lse.x
            n = con.lse.n
            b = con.ub

            P = Problem()

            u = P.add_variable("__u", n)
            P.add_constraint((u | 1) <= 1)

            for i in range(n):
                P.add_constraint((u[i] // 1 // (x[i] - b)) << ExponentialCone())

            return P

        @classmethod
        def dual(cls, auxVarPrimals, auxConDuals, options):
            """Implement :meth:`~.constraint.ConstraintConversion.dual`."""
            # TODO: Verify that this is the dual.
            return auxConDuals[0]

    def __init__(self, lse, upperBound):
        """Construct a :class:`LogSumExpConstraint`.

        :param ~picos.expressions.LogSumExp lse:
            Constrained expression.
        :param ~picos.expressions.AffineExpression upperBound:
            Upper bound on the expression.
        """
        from ..expressions import AffineExpression, LogSumExp

        assert isinstance(lse, LogSumExp)
        assert isinstance(upperBound, AffineExpression)
        assert len(upperBound) == 1

        self.lse = lse
        self.ub  = upperBound

        super(LogSumExpConstraint, self).__init__(
            lse._typeStr if isinstance(lse, LogSumExp) else "LSE")

    @property
    def exponents(self):
        """The affine exponents of the bounded log-sum-exp expression."""
        return self.lse.x

    @cached_property
    def le0(self):
        """The :class:`~.exp_logsumexp.LogSumExp` posed to be at most zero."""
        from ..expressions import LogSumExp

        if self.ub.is0:
            return self.lse
        else:
            return LogSumExp(self.lse.x - self.ub)

    Subtype = namedtuple("Subtype", ("argdim",))

    def _subtype(self):
        return self.Subtype(self.lse.n)

    @classmethod
    def _cost(cls, subtype):
        return subtype.argdim + 1

    def _expression_names(self):
        yield "lse"
        yield "ub"

    def _str(self):
        return glyphs.le(self.lse.string, self.ub.string)

    def _get_size(self):
        return (1, 1)

    def _get_slack(self):
        return self.ub.value - self.lse.value


# --------------------------------------
__all__ = api_end(_API_START, globals())



# coding: utf-8

# ------------------------------------------------------------------------------
# Copyright (C) 2019 Maximilian Stahlberg
# Based on the original picos.expressions module by Guillaume Sagnol.
#
# This file is part of PICOS.
#
# PICOS is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PICOS is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# ------------------------------------------------------------------------------

"""Implements :class:`SumExtremes`."""

import operator
from collections import namedtuple

import cvxopt
import numpy

from .. import glyphs
from ..apidoc import api_end, api_start
from ..caching import cached_unary_operator
from ..constraints import Constraint, SumExtremesConstraint
from .data import convert_and_refine_arguments, convert_operands, cvx2np
from .exp_affine import AffineExpression, ComplexAffineExpression
from .expression import Expression, refine_operands, validate_prediction

_API_START = api_start(globals())
# -------------------------------


class SumExtremes(Expression):
    r"""Sum of the :math:`k` largest or smallest elements or eigenvalues.

    :Definition:

    Let :math:`k \in \mathbb{Z}_{\geq 1}`.

    1.  If :math:`x` is an :math:`n`-dimensional real vector or matrix and
        ``eigenvalues == False``, then this is the sum of the :math:`k \leq n`
        largest or smallest scalar elements of :math:`x`, depending on the truth
        value of ``largest``.

        Special cases:

        -   If :math:`k = 1`, this is either the largest element
            :math:`\max_{i = 1}^n \operatorname{vec}(x)_i` or the smallest
            element :math:`\min_{i = 1}^n \operatorname{vec}(x)_i` of :math:`x`.
        -   If :math:`k = n`, this is the sum of all elements
            :math:`\langle x, 1 \rangle` of :math:`x`.

    2.  If :math:`X` is an :math:`n \times n` hermitian matrix and
        ``eigenvalues == True``, then this is the sum of the :math:`k \leq n`
        largest or smallest eigenvalues of :math:`X`, depending on the truth
        value of ``largest``. Recall that the eigenvalues of a hermitian matrix
        are real.

        Special cases:

        -   If :math:`k = 1`, this is either the largest eigenvalue
            :math:`\lambda_{\max}(X)` or the smallest eigenvalue
            :math:`\lambda_{\min}(X)` of :math:`X`.
        -   If :math:`k = n`, this equals the trace
            :math:`\operatorname{tr}(X)`.

    If the given :math:`k` exceeds the :math:`n` of either case, then :math:`k`
    is silently clipped to :math:`n`.
    """

    # --------------------------------------------------------------------------
    # Initialization and factory methods.
    # --------------------------------------------------------------------------

    @convert_and_refine_arguments("x")
    def __init__(self, x, k, largest, eigenvalues=False):
        """Construct a :class:`SumExtremes`.

        :param x: The affine expression to take a sum over.
        :type x: ~picos.expressions.ComplexAffineExpression
        :param int k: Number of summands.
        :param bool largest: Whether to sum over the largest (eigen)values as
            opposed to the smallest.
        :param bool eigenvalues: Whether to sum eigenvalues instead of elements.
        """
        largest     = bool(largest)
        eigenvalues = bool(eigenvalues)

        lStr = "largest" if largest else "smallest"
        eStr = "eigenvalues" if eigenvalues else "scalar elements"
        what = "{} {}".format(lStr, eStr)

        # Validate x.
        if not isinstance(x, ComplexAffineExpression):
            raise TypeError("Can only sum {} of an affine expression, not of "
                "{}.".format(what, type(x).__name__))

        # Further validate x.
        if eigenvalues:
            if not x.square:
                raise TypeError("Cannot sum {} of {} as its shape of {} is not "
                    "square.".format(what, x.string, glyphs.shape(x.shape)))
            elif not x.hermitian:
                raise NotImplementedError(
                    "Summing the {0} of {1} is not supported as {1} is not "
                    "necessarily hermitian.".format(what, x.string))
        else:
            if not isinstance(x, AffineExpression):
                raise TypeError("Can only sum {} of a real-valued expression "
                    "but {} is properly complex.".format(what, x.string))

        # Validate k.
        if int(k) != k:
            raise ValueError(
                "Conversion of k = {} to an integer is ambiguous.".format(k))
        k = int(k)
        if k < 1:
            raise ValueError(
                "Number of {} to sum must be positive.".format(what))

        # Clip k to be at most n.
        k = min(k, x.shape[0]) if eigenvalues else min(k, len(x))

        # Find out if all (eigen)values are summed.
        full = k == x.shape[0] if eigenvalues else k == len(x)
        assert len(x) != 1 or full

        self._x           = x
        self._k           = k
        self._largest     = largest
        self._eigenvalues = eigenvalues
        self._full        = full

        s, lbd = x.string, glyphs.lambda_()
        if full:
            if eigenvalues:
                typeStr = "Sum of Eigenvalues"
                symbStr = symbStr = glyphs.trace(s)
            else:
                typeStr = "Sum of Elements"
                symbStr = glyphs.sum(s)
        elif k > 1:
            if eigenvalues and largest:
                typeStr = "Sum of Largest Eigenvalues"
                symbStr = glyphs.make_function(
                    "sum_{}_largest_{}".format(k, lbd))(s)
            elif eigenvalues and not largest:
                typeStr = "Sum of Smallest Eigenvalues"
                symbStr = glyphs.make_function(
                    "sum_{}_smallest_{}".format(k, lbd))(s)
            elif not eigenvalues and largest:
                typeStr = "Sum of Largest Elements"
                symbStr = glyphs.make_function("sum_{}_largest".format(k))(s)
            else:
                typeStr = "Sum of Smallest Elements"
                symbStr = glyphs.make_function("sum_{}_smallest".format(k))(s)
        else:
            if eigenvalues and largest:
                typeStr = "Largest Eigenvalue"
                symbStr = glyphs.make_function("{}_max".format(lbd))(s)
            elif eigenvalues and not largest:
                typeStr = "Smallest Eigenvalue"
                symbStr = glyphs.make_function("{}_min".format(lbd))(s)
            elif not eigenvalues and largest:
                typeStr = "Largest Element"
                symbStr = glyphs.max(s)
            else:
                typeStr = "Smallest Element"
                symbStr = glyphs.min(s)

        Expression.__init__(self, typeStr, symbStr)

    # --------------------------------------------------------------------------
    # Abstract method implementations and method overridings, except _predict.
    # --------------------------------------------------------------------------

    def _get_refined(self):
        if self._x.constant:
            return AffineExpression.from_constant(self.value, 1, self._symbStr)
        elif self._full:
            if len(self._x) == 1:
                return self._x  # Don't carry the string for an identity.
            if self._eigenvalues:
                return self._x.tr  # Symbolic strings already match.
            else:
                return (1 | self._x).renamed(self._symbStr)
        else:
            return self

    Subtype = namedtuple("Subtype",
        ("argdim", "k", "largest", "eigenvalues", "complex"))

    def _get_subtype(self):
        return self.Subtype(len(self._x), self._k, self._largest,
            self._eigenvalues, self._x.complex)

    def _get_value(self):
        value = self._x._get_value()

        if self._eigenvalues:
            value = sorted(numpy.linalg.eigvalsh(cvx2np(value)))
        else:
            value = sorted(value)

        value = sum(value[-self._k:] if self._largest else value[:self._k])
        value = cvxopt.matrix(value)

        return value

    @cached_unary_operator
    def _get_variables(self):
        return self._x.variables

    def _is_convex(self):
        return self._largest or self._full

    def _is_concave(self):
        return not self._largest or self._full

    def _replace_variables(self, var_map):
        return self.__class__(self._x._replace_variables(var_map),
            self._k, self._largest, self._eigenvalues)

    # --------------------------------------------------------------------------
    # Python special method implementations, except constraint-creating ones.
    # --------------------------------------------------------------------------

    @convert_operands(scalarRHS=True)
    @refine_operands()
    def __mul__(self, other):
        if isinstance(other, AffineExpression):
            if not other.constant:
                raise NotImplementedError("You may only multiply a nonconstant "
                    "PICOS sum of extremes with a constant term.")

            if other.value < 0:
                raise NotImplementedError("You may only multiply a nonconstant "
                    "PICOS sum of extremes with a nonnegative term.")

            product = SumExtremes(
                other.value*self._x, self._k, self._largest, self._eigenvalues)
            product._typeStr = "Scaled " + product._typeStr
            product._symbStr = glyphs.clever_mul(self.string, other.string)
            return product
        else:
            return NotImplemented

    @convert_operands(scalarRHS=True)
    @refine_operands()
    def __rmul__(self, other):
        if isinstance(other, AffineExpression):
            product = self.__mul__(other)
            # NOTE: __mul__ always creates a fresh expression.
            product._symbStr = glyphs.clever_mul(other.string, self.string)
            return product
        else:
            return NotImplemented

    # --------------------------------------------------------------------------
    # Methods and properties that return expressions.
    # --------------------------------------------------------------------------

    @property
    def x(self):
        """The expression under the sum."""
        return self._x

    # --------------------------------------------------------------------------
    # Methods and properties that describe the expression.
    # --------------------------------------------------------------------------

    @property
    def k(self):
        """Number of (eigen)values to sum."""
        return self._k

    @property
    def largest(self):
        """Whether the sum concerns largest values as opposed to smallest."""
        return self._largest

    @property
    def eigenvalues(self):
        """Whether the sum concerns eigenvalues as opposed to elements."""
        return self._eigenvalues

    @property
    def full(self):
        """Whether the sum concerns *all* (eigen)values of the expression."""
        return self._full

    # --------------------------------------------------------------------------
    # Constraint-creating operators, and _predict.
    # --------------------------------------------------------------------------

    @classmethod
    def _predict(cls, subtype, relation, other):
        assert isinstance(subtype, cls.Subtype)

        n = subtype.argdim
        k = subtype.k
        e = subtype.eigenvalues
        c = subtype.complex

        kmax = int(n**0.5) if e else n
        full = k == kmax

        convex  = subtype.largest or full
        concave = not subtype.largest or full

        if relation == operator.__le__:
            if not convex:
                return NotImplemented

            if issubclass(other.clstype, AffineExpression) \
            and other.subtype.dim == 1:
                return SumExtremesConstraint.make_type(n, k, e, c)
        elif relation == operator.__ge__:
            if not concave:
                return NotImplemented

            if issubclass(other.clstype, AffineExpression) \
            and other.subtype.dim == 1:
                return SumExtremesConstraint.make_type(n, k, e, c)

        return NotImplemented

    @convert_operands(scalarRHS=True)
    @validate_prediction
    @refine_operands()
    def __le__(self, other):
        if not self.convex:
            raise TypeError("Cannot upper-bound the nonconvex expression {}."
                .format(self._symbStr))

        if isinstance(other, AffineExpression):
            return SumExtremesConstraint(self, Constraint.LE, other)
        else:
            return NotImplemented

    @convert_operands(scalarRHS=True)
    @validate_prediction
    @refine_operands()
    def __ge__(self, other):
        if not self.concave:
            raise TypeError("Cannot upper-bound the nonconcave expression {}."
                .format(self._symbStr))

        if isinstance(other, AffineExpression):
            return SumExtremesConstraint(self, Constraint.GE, other)
        else:
            return NotImplemented


# --------------------------------------
__all__ = api_end(_API_START, globals())
